"""
Plot 1D entropy in the time or frequency domain
"""

import numpy as np
import scipy.signal as signal
import scipy.fft
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from quantum_inferno import info


def plot_fft_compare(sig_in,
                     fs_hz: float,
                     duration_no_pad_points: int,
                     welch_points: int,
                     overlap_points: int,
                     fmin_hz: float = 2E-5,
                     fmax_hz: float = 1E-2,
                     station_str: str = None,
                     do_show: bool = True,
                     do_save: bool = False,
                     save_path: str = None):
    """
    Comparisons for time and fft isnr and esnr
    :param sig_in:
    :param fs_hz:
    :param duration_no_pad_points:
    :param welch_points:
    :param overlap_points:
    :param fmin_hz: Plotting min frequency limit
    :param fmax_hz: Plotting max frequency limit
    :param station_str: Station ID
    :param do_show:
    :param do_save:
    :param save_path:
    :return:
    """

    time_idx = duration_no_pad_points
    # TODO: Clean up, this is silly
    # TODO: Build the distribution only from the portion of the spectrum that was used
    sig_in_no_pad = sig_in[0:time_idx]
    time = np.arange(len(sig_in_no_pad))/fs_hz/3600

    [tdr_sig, _, _, _, tdr_isnr, tdr_esnr] = \
        info.shannon_tdr(sig_in_real=sig_in_no_pad)

    rfft_sig1 = scipy.fft.rfft(x=sig_in_no_pad)
    rfft_freq1 = scipy.fft.rfftfreq(n=len(sig_in_no_pad), d=1/fs_hz)
    [fft_marginal1, fft_angle_rads1, _, _, fft_isnr1, fft_esnr1] = \
        info.shannon_fft(fft_sig=rfft_sig1)

    rfft_sig2 = scipy.fft.rfft(x=sig_in)
    rfft_freq2 = scipy.fft.rfftfreq(n=len(sig_in), d=1/fs_hz)
    [fft_marginal2, fft_angle_rads2, _, _, fft_isnr2, fft_esnr2] = \
        info.shannon_fft(fft_sig=rfft_sig2)

    # Use zero padding
    rfft_freq3, welch_power = scipy.signal.welch(x=sig_in_no_pad, fs=fs_hz,
                                                 nperseg=welch_points, noverlap=overlap_points,
                                                 nfft=len(sig_in))
    [fft_marginal3, _, _, _, fft_isnr3, fft_esnr3] = \
        info.shannon_fft(fft_sig=np.sqrt(welch_power))

    # set the font size
    fontsz = 'x-large'
    # Initialize subplots
    fig, ax = plt.subplots(3, 2)
    fig.set_size_inches(16, 8)

    # Axis numbering line
    ref_line = mlines.Line2D([], [])

    ax[2, 0].plot(time, tdr_sig)
    ax[2, 0].set_ylabel('Norm Signal', fontsize=fontsz)
    ax[2, 0].set_xlabel('Time, hours', fontsize=fontsz)
    ax[2, 0].set_xlim(time[0], time[time_idx-1])
    ax[2, 0].tick_params(labelsize='x-large')
    ax[2, 0].grid(True)
    if station_str is None:
        label_a = '(a)'
    else:
        label_a = '(a)' + station_str

    ax[2, 0].legend(handles=[ref_line], labels=[label_a], loc=2,
                    frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')

    ax[1, 0].plot(time, tdr_isnr, '.')
    ax[1, 0].set_xticklabels([])
    ax[1, 0].set_ylim(-8, 8)
    ax[1, 0].set_xlim(time[0], time[time_idx-1])
    ax[1, 0].tick_params(labelsize='x-large')
    ax[1, 0].set_ylabel('Information SNR', fontsize=fontsz)
    ax[1, 0].grid(True)
    ax[1, 0].legend(handles=[ref_line], labels=['(b)'], loc=2,
                    frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')

    ax[0, 0].plot(time, tdr_esnr, '.')
    ax[0, 0].set_xticklabels([])
    ax[0, 0].set_ylabel('Entropy SNR', fontsize=fontsz)
    ax[0, 0].set_xlim(time[0], time[time_idx-1])
    ax[0, 0].tick_params(labelsize='x-large')
    ax[0, 0].grid(True)
    ax[0, 0].legend(handles=[ref_line], labels=['(c)'], loc=2,
                    frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')

    # Phase can be a bit boring
    # ax[2, 1].semilogx(rfft_freq1, fft_angle_rads1, '.', base=10, label='FFT')
    # ax[2, 1].semilogx(rfft_freq2, fft_angle_rads2, '.', base=10, label='FFT Padded')
    # ax[2, 1].set_xlim(fmin_hz, fmax)
    # # ax[2, 1].set_ylim(angle_min, angle_max+10)
    # ax[2, 1].set_ylim(-500, 50)
    # ax[2, 1].set_ylabel('Phase, rad', fontsize=fontsz)
    # ax[2, 1].set_xlabel('Frequency, Hz', fontsize=fontsz)
    # ax[2, 1].tick_params(labelsize='x-large')
    # ax[2, 1].grid(True)
    # panel_number = ax[2, 1].legend(handles=[ref_line], labels=['(d)'], loc=2,
    #                                frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')
    # ax[2, 1].legend(loc=1)
    # ax[2, 1].add_artist(panel_number)

    ax[2, 1].semilogx(rfft_freq1, fft_marginal1, '--', base=10, label='FFT')
    ax[2, 1].semilogx(rfft_freq2, fft_marginal2, '.', base=10, label='FFT Padded')
    ax[2, 1].semilogx(rfft_freq3, fft_marginal3, '.', base=10, label='Welch Padded')
    ax[2, 1].set_xlim(fmin_hz, fmax_hz)
    ax[2, 1].set_ylabel('Marginals', fontsize=fontsz)
    ax[2, 1].set_xlabel('Frequency, Hz', fontsize=fontsz)
    ax[2, 1].tick_params(labelsize='x-large')
    ax[2, 1].grid(True)
    panel_number = ax[2, 1].legend(handles=[ref_line], labels=['(d)'], loc=2,
                                   frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')
    ax[2, 1].legend(loc=1)
    ax[2, 1].add_artist(panel_number)

    ax[1, 1].semilogx(rfft_freq1, fft_isnr1, '--', base=10, label='FFT')
    ax[1, 1].semilogx(rfft_freq2, fft_isnr2, '.', base=10, label='FFT Padded')
    ax[1, 1].semilogx(rfft_freq3, fft_isnr3, '.', base=10, label='Welch Padded')
    ax[1, 1].set_xticklabels([])
    ax[1, 1].set_ylabel('Information SNR', fontsize=fontsz)
    ax[1, 1].set_ylim(-4, 12)
    ax[1, 1].set_xlim(fmin_hz, fmax_hz)
    ax[1, 1].tick_params(labelsize='x-large')
    ax[1, 1].grid(True)
    panel_number = ax[1, 1].legend(handles=[ref_line], labels=['(e)'], loc=2,
                                   frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')
    ax[1, 1].legend(loc=1)
    ax[1, 1].add_artist(panel_number)

    ax[0, 1].semilogx(rfft_freq1, fft_esnr1, '.', base=10, label='FFT')
    ax[0, 1].semilogx(rfft_freq2, fft_esnr2, '.', base=10, label='FFT Padded')
    ax[0, 1].semilogx(rfft_freq3, fft_esnr3, '.', base=10, label='Welch Padded')
    # NOTE: Below returns isnr because log2(plog2(p)) = log2(p) + log2(log2(p)) ~ log2p
    # ax[0, 1].semilogx(rfft_freq1, np.log2(fft_esnr1), '--', base=10, label='FFT')
    # ax[0, 1].semilogx(rfft_freq2, np.log2(fft_esnr2), '.', base=10, label='FFT Padded')
    # ax[0, 1].semilogx(rfft_freq3, np.log2(fft_esnr3), '.', base=10, label='Welch Padded')
    # ax[0, 1].set_ylim(-4, 12)
    ax[0, 1].set_xticklabels([])
    ax[0, 1].set_ylabel('Entropy SNR', fontsize=fontsz)
    # ax[0, 1].set_title(title_label)
    ax[0, 1].set_xlim(fmin_hz, fmax_hz)
    ax[0, 1].tick_params(labelsize='x-large')
    ax[0, 1].grid(True)
    panel_number = ax[0, 1].legend(handles=[ref_line], labels=['(f)'], loc=2,
                                   frameon=False, handlelength=0, handletextpad=-0.5, fontsize='x-large')
    ax[0, 1].legend(loc=1)
    ax[0, 1].add_artist(panel_number)
    plt.tight_layout()

    if do_save:
        if save_path is None:
            print('Save path must be specified! Exiting program.')
            exit()
        fig.savefig(save_path)

    if do_show:
        plt.show()


"""
This module contains general utilities that can work with values containing nans.
TODO: Break up into smaller modules
"""

import os
from enum import Enum
from typing import Tuple, Union
import numpy as np
from scipy import interpolate, signal
from scipy.integrate import cumulative_trapezoid
from quantum_inferno.scales_dyadic import EPSILON64 as EPSILON

# from redvox.common import date_time_utils as dt

""" Directory check & make """


def checkmake_dir(DIR_NAME):
    # Check export figure folder exist; if not, make
    existing_dir: bool = os.path.isdir(DIR_NAME)
    if not existing_dir:
        os.makedirs(DIR_NAME)


""" Logical power of two flag """


def is_power_of_two(n):
    """ Returns not if not positive and a power of two """
    return not (n > 0 and (n & (n - 1)))


""" Time/Sample Duration Utils """


def duration_points(sample_rate_hz: float, time_s: float) -> Tuple[int, int, int]:
    """
    Compute number of points
    :param sample_rate_hz: sample rate in Hz
    :param time_s: time scale, period or duration
    :return: number of points, floor and ceiling of log2 of number of points
    """
    points_float: float = sample_rate_hz * time_s
    points_int: int = int(points_float)
    points_floor_log2: int = int(np.floor(np.log2(points_float)))
    points_ceil_log2: int = int(np.ceil(np.log2(points_float)))

    return points_int, points_floor_log2, points_ceil_log2


def duration_ceil(sample_rate_hz: float, time_s: float) -> Tuple[int, int, float]:
    """
    Compute ceiling of the number of points, and convert to seconds
    :param sample_rate_hz: sample rate in Hz
    :param time_s: time scale, period or duration
    :return: ceil of log 2 of number of points, power of two number of points, corresponding time in s
    """
    points_float: float = sample_rate_hz * time_s
    points_ceil_log2: int = int(np.ceil(np.log2(points_float)))
    points_ceil_pow2: int = 2**points_ceil_log2
    time_ceil_pow2_s: float = points_ceil_pow2 / sample_rate_hz

    return points_ceil_log2, points_ceil_pow2, time_ceil_pow2_s


def duration_floor(sample_rate_hz: float, time_s: float) -> Tuple[int, int, float]:
    """
    Compute floor of the number of points, and convert to seconds
    :param sample_rate_hz: sample rate in Hz
    :param time_s: time scale, period or duration
    :return: floor of log 2 of number of points, power of two number of points, corresponding time in s
    """
    points_float: float = sample_rate_hz * time_s
    points_floor_log2: int = int(np.floor(np.log2(points_float)))
    points_floor_pow2: int = 2**points_floor_log2
    time_floor_pow2_s: float = points_floor_pow2 / sample_rate_hz

    return points_floor_log2, points_floor_pow2, time_floor_pow2_s


""" Sampling Utils """


def resampled_mean_no_hop(sig_wf: np.array,
                          sig_time: np.array,
                          resampling_factor: int):

    """
    Resamples a time series
    Assume len(sig_time) is a power of two
    Assume len(resampling_factor) is a power of two
    :param sig_wf: audio waveform
    :param sig_time: audio timestamps in seconds, same dimensions as sig
    :param power_tfr: time-frequency representation with same number of columns as sig
    :param resampling_factor: downsampling factor

    :return: rms_sig_wf, rms_sig_time_s
    """

    if resampling_factor <= 1:
        print('No Resampling/Averaging!')
        exit()
    # Variance over the signal waveform, all bands

    points_resampled: int = int(np.round((len(sig_wf)/resampling_factor)))
    new_wf_resampled = np.reshape(a=sig_wf, newshape=(points_resampled, resampling_factor))

    new_sig_wf_mean = np.mean(new_wf_resampled, axis=1)

    # Resampled signal time
    new_sig_time_s = sig_time[0::resampling_factor]

    return [new_sig_time_s, new_sig_wf_mean]


def resampled_power_per_band_no_hop(sig_wf: np.array,
                                    sig_time: np.array,
                                    power_tfr: np.array,
                                    resampling_factor: int):

    """
    Resamples a time series and the wavelet tfr time dimension
    Assume len(sig_time) is a power of two
    Assume len(resampling_factor) is a power of two
    :param sig_wf: audio waveform
    :param sig_time: audio timestamps in seconds, same dimensions as sig
    :param power_tfr: time-frequency representation with same number of columns (time) as sig
    :param resampling_factor: downsampling factor

    :return: rms_sig_wf, rms_sig_time_s
    """

    var_sig = (sig_wf - np.mean(sig_wf))**2

    if resampling_factor <= 1:
        print('No Resampling/Averaging!')
        exit()
    # Variance over the signal waveform, all bands
    # Number of rows (frequency)
    number_bands = power_tfr.shape[0]
    points_resampled: int = int(np.round((len(sig_wf)/resampling_factor)))
    var_wf_resampled = np.reshape(a=var_sig, newshape=(points_resampled, resampling_factor))

    var_sig_wf_mean = np.mean(var_wf_resampled, axis=1)
    var_sig_wf_max = np.max(var_wf_resampled, axis=1)
    var_sig_wf_min = np.min(var_wf_resampled, axis=1)

    # Reshape TFR
    var_tfr_resampled = np.reshape(a=power_tfr, newshape=(number_bands, points_resampled, resampling_factor))
    print('var_tfr_resampled.shape')
    print(var_tfr_resampled.shape)
    var_tfr_mean = np.mean(var_tfr_resampled, axis=2)
    var_tfr_max = np.max(var_tfr_resampled, axis=2)
    var_tfr_min = np.min(var_tfr_resampled, axis=2)

    # Resampled signal time
    var_sig_time_s = sig_time[0::resampling_factor]

    return [var_sig_time_s, var_sig_wf_mean, var_sig_wf_max, var_sig_wf_min, var_tfr_mean, var_tfr_max, var_tfr_min]


def resample_uneven_signal(sig_wf: np.ndarray,
                           sig_epoch_s: np.ndarray,
                           sample_rate_new_hz: float = None):
    """

    :param sig_wf:
    :param sig_epoch_s:
    :param sample_rate_new_hz:
    :return:
    """

    if sample_rate_new_hz is None:
        interval_from_epoch_s = np.mean(np.diff(sig_epoch_s))
        # Round up
        sample_rate_new_hz = np.ceil(1/interval_from_epoch_s)

    interval_s = 1/sample_rate_new_hz
    sig_new_epoch_s = np.arange(sig_epoch_s[0], sig_epoch_s[-1], interval_s)
    f = interpolate.interp1d(sig_epoch_s, sig_wf)
    sig_new_wf = f(sig_new_epoch_s)
    return sig_new_wf, sig_new_epoch_s


def upsample_fourier(sig_wf: np.ndarray,
                     sig_sample_rate_hz: float,
                     new_sample_rate_hz: float = 8000.) -> np.ndarray:
    """
    Upsample the Fourier way.

    :param sig_wf: input signal waveform, reasonably well preprocessed
    :param sig_sample_rate_hz: signal sample rate
    :param new_sample_rate_hz: resampling sample rate
    :return: resampled signal
    """
    sig_len = len(sig_wf)
    new_len = int(sig_len * new_sample_rate_hz / sig_sample_rate_hz)
    sig_resampled = signal.resample(x=sig_wf, num=new_len)
    return sig_resampled


def taper_tukey(sig_wf_or_time: np.ndarray,
                fraction_cosine: float) -> np.ndarray:
    """
    Constructs a symmetric Tukey window with the same dimensions as a time or signal numpy array.
    fraction_cosine = 0 is a rectangular window, 1 is a Hann window

    :param sig_wf_or_time: input signal or time
    :param fraction_cosine: fraction of the window inside the cosine tapered window, shared between the head and tail
    :return: tukey taper window amplitude
    """
    return signal.windows.tukey(M=np.size(sig_wf_or_time), alpha=fraction_cosine, sym=True)


# Integrals and derivatives
def integrate_cumtrapz(timestamps_s: np.ndarray,
                       sensor_wf: np.ndarray,
                       initial_value: float = 0) -> np.ndarray:
    """
    Cumulative trapazoid integration using scipy.integrate.cumulative_trapezoid
    Initiated by Kei 2106, work in progress. See blast_derivative_integral for validation.

    :param timestamps_s: timestamps corresponding to the data in seconds
    :param sensor_wf: data to integrate using cumulative trapezoid
    :param initial_value: the value to add in the initial of the integrated data to match length of input (default is 0)
    :return: integrated data with the same length as the input
    """
    integrated_data = cumulative_trapezoid(x=timestamps_s,
                                           y=sensor_wf,
                                           initial=initial_value)
    return integrated_data


def derivative_gradient(timestamps_s: np.ndarray,
                        sensor_wf: np.ndarray) -> np.ndarray:
    """
    Derivative using gradient

    :param timestamps_s: timestamps corresponding to the data in seconds
    :param sensor_wf: data to integrate using cumulative trapezoid
    :return: derivative data with the same length as the input
    """
    # derivative_data = np.gradient(sensor_wf)/np.gradient(timestamps_s)
    derivative_data = np.gradient(sensor_wf, timestamps_s)

    return derivative_data


def derivative_diff(timestamps_s: np.ndarray,
                    sensor_wf: np.ndarray) -> np.ndarray:
    """
    Derivative using diff

    :param timestamps_s: timestamps corresponding to the data in seconds
    :param sensor_wf: data to integrate using cumulative trapezoid
    :return: derivative data with the same length as the input. Hold/repeat last value
    """

    derivative_data0 = np.diff(sensor_wf)/np.diff(timestamps_s)
    derivative_data = np.append(derivative_data0, derivative_data0[-1])

    return derivative_data


class ExtractionType(Enum):
    """
    Enumeration of valid extraction types.

    ARGMAX = max of the absolute value of the signal
    SIGMAX = fancier signal picker, from POSITIVE max
    BITMAX = fancier signal picker, from ABSOLUTE max
    """
    ARGMAX: str = "argmax"
    SIGMAX: str = "sigmax"
    BITMAX: str = "bitmax"


def sig_extract(sig: np.ndarray,
                time_epoch_s: np.ndarray,
                intro_s: float,
                outro_s: float,
                pick_bits_below_max: float = 1.,
                pick_time_interval_s: float = 1.,
                extract_type: ExtractionType = ExtractionType.ARGMAX) -> Tuple[np.ndarray, np.ndarray, float]:
    """
    Extract signal and time relative to reference index

    :param sig: input signal
    :param time_epoch_s: signal epoch time
    :param intro_s: time before pick
    :param outro_s: time after pick
    :param pick_bits_below_max: pick treshold in bits below max
    :param pick_time_interval_s: pick time interval between adjacent max point
    :param extract_type: Type of extraction, see ExtractionType class
    :return: extracted signal np.ndarray, extracted signal timestamps np.ndarray, and pick time
    """

    sig_sample_interval_s = np.mean(np.diff(time_epoch_s))

    if extract_type == ExtractionType.ARGMAX:
        index_max = np.argmax(np.abs(sig))
        # Max pick
        pick_time_epoch_s = time_epoch_s[index_max]
    elif extract_type == ExtractionType.SIGMAX:
        time_index_pick_all = \
            picker_signal_max_index(sig=sig, sig_sample_rate_hz=1./sig_sample_interval_s,
                                    bits_pick=pick_bits_below_max, time_interval_s=pick_time_interval_s)
        # First pick
        pick_time_epoch_s = time_epoch_s[time_index_pick_all[0]]
    elif extract_type == ExtractionType.BITMAX:
        time_index_pick_all = \
            picker_signal_bit_index(sig=sig, sig_sample_rate_hz=1./sig_sample_interval_s,
                                    bits_pick=pick_bits_below_max, time_interval_s=pick_time_interval_s)
        # First pick
        pick_time_epoch_s = time_epoch_s[time_index_pick_all[0]]
    else:
        print('Unexpected extraction type to sig_extract, return max')
        index_max = np.argmax(np.abs(sig))
        # Max pick
        pick_time_epoch_s = time_epoch_s[index_max]

    epoch_s_start = pick_time_epoch_s - intro_s
    epoch_s_stop = pick_time_epoch_s + outro_s

    intro_index = np.argmin(np.abs(time_epoch_s - epoch_s_start))
    outro_index = np.argmin(np.abs(time_epoch_s - epoch_s_stop))
    sig_wf = sig[intro_index: outro_index]
    sig_epoch_s = time_epoch_s[intro_index: outro_index]

    return sig_wf, sig_epoch_s, pick_time_epoch_s


def sig_frame(sig: np.ndarray,
              time_epoch_s: np.ndarray,
              epoch_s_start: float,
              epoch_s_stop: float) -> Tuple[np.ndarray, np.ndarray]:
    """
    Frame one-component signal within start and stop epoch times

    :param sig: input signal
    :param time_epoch_s: input epoch time in seconds
    :param epoch_s_start: start epoch time
    :param epoch_s_stop: stop epoch time
    :return: truncated time series and time
    """
    intro_index = np.argmin(np.abs(time_epoch_s - epoch_s_start))
    outro_index = np.argmin(np.abs(time_epoch_s - epoch_s_stop))
    sig_wf = sig[intro_index: outro_index]
    sig_epoch_s = time_epoch_s[intro_index: outro_index]

    return sig_wf, sig_epoch_s


def sig3c_frame(sig3c: np.ndarray,
                time_epoch_s: np.ndarray,
                epoch_s_start: float,
                epoch_s_stop: float) -> Tuple[np.ndarray, np.ndarray]:
    """
    Frame three-component signal within start and stop epoch times

    :param sig3c: input signal with three components
    :param time_epoch_s: input epoch time in seconds
    :param epoch_s_start: start epoch time
    :param epoch_s_stop: stop epoch time
    :return: truncated time series and time
    """
    intro_index = np.argmin(np.abs(time_epoch_s - epoch_s_start))
    outro_index = np.argmin(np.abs(time_epoch_s - epoch_s_stop))
    sig_wf = sig3c[:, intro_index: outro_index]
    sig_epoch_s = time_epoch_s[intro_index: outro_index]

    return sig_wf, sig_epoch_s


def dbepsilon(x: np.ndarray) -> np.ndarray:
    """
    Converts the absolute value of a time series to dB

    :param x: time series
    :return: ndarray
    """
    y = 10 * np.log10(np.abs(x ** 2) + EPSILON)
    return y


def dbepsilon_max(x: np.ndarray) -> float:
    """
    Returns the max of the absolute value of a time series to dB

    :param x: time series
    :return: float
    """
    y = 10 * np.log10(np.max(np.abs(x ** 2)) + EPSILON)
    return y


def log2epsilon(x: np.ndarray) -> np.ndarray:
    """
    log 2 of the absolute value of linear amplitude, with EPSILON to avoid singularities

    :param x: time series or fft - not power
    :return: ndarray
    """
    y = np.log2(np.abs(x) + EPSILON)
    return y


def log2epsilon_max(x: np.ndarray) -> float:
    """
    max of the log 2 of absolute value of linear amplitude, with EPSILON to avoid singularities

    :param x: time series or fft - not power
    :return: float
    """
    y = np.max(np.log2(np.abs(x) + EPSILON))
    return y


"""
Picker modules
"""


def picker_signal_max_index(sig: np.array,
                            sig_sample_rate_hz: float,
                            bits_pick: float,
                            time_interval_s: float) -> np.array:
    """
    Finds the picker index for the POSITIVE max of a signal

    :param sig: array of waveform data
    :param sig_sample_rate_hz: float sample rate of reference sensor
    :param bits_pick: detection threshold in bits loss
    :param time_interval_s: min time interval between events
    :return: picker index
    """

    # Compute the distance
    distance_points = int(time_interval_s * sig_sample_rate_hz)
    height_min = np.max(sig) - 2 ** bits_pick
    time_index_pick, _ = signal.find_peaks(sig, height=height_min, distance=distance_points)

    return time_index_pick


def picker_signal_bit_index(sig: np.array,
                            sig_sample_rate_hz: float,
                            bits_pick: float,
                            time_interval_s: float) -> np.ndarray:
    """
    Finds the picker index from the ABSOLUTE max in bits

    :param sig: array of waveform data
    :param sig_sample_rate_hz: float sample rate of reference sensor
    :param bits_pick: detection treshold in db loss
    :param time_interval_s: min time interval between events
    :return: picker index
    """

    # Compute the distance
    distance_points = int(time_interval_s * sig_sample_rate_hz)
    sig_bit = log2epsilon(sig)
    height_min = log2epsilon_max(sig) - bits_pick
    index_pick, _ = signal.find_peaks(sig_bit, height=height_min, distance=distance_points)

    return index_pick


def picker_comb(sig_pick: np.ndarray,
                index_pick: np.ndarray) -> np.ndarray:
    """
    Constructs a comb function from the picks

    :param sig_pick: 1D record corresponding to the picks
    :param index_pick: indexes for the picks
    :return: comb with unit amplitude
    """
    comb = np.zeros(sig_pick.shape)
    comb[index_pick] = np.ones(index_pick.shape)
    return comb


"""
Matrix transformations
"""


def sum_columns(sxx: np.ndarray) -> np.ndarray:
    """
    Sum over all the columns in a 1D or 2D array

    :param sxx: input vector or matrix
    :return: ndarray with sum
    """
    if not isinstance(sxx, np.ndarray):
        raise TypeError('Input must be array.')
    elif len(sxx) == 0:
        raise ValueError('Cannot compute on empty array.')

    if np.isnan(sxx).any() or np.isinf(sxx).any():
        sxx = np.nan_to_num(sxx)

    if len(sxx.shape) == 1:
        sum_c = np.sum(sxx, axis=0)
    elif len(sxx.shape) == 2:
        sum_c = np.sum(sxx, axis=1)
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(sxx.shape)))

    return sum_c


def mean_columns(sxx: np.ndarray) -> np.ndarray:
    """
    Compute the mean of the columns in a 1D or 2D array

    :param sxx: input vector or matrix
    :return: ndarray with mean
    """
    if not isinstance(sxx, np.ndarray):
        raise TypeError('Input must be array.')
    elif len(sxx) == 0:
        raise ValueError('Cannot compute on empty array.')

    if np.isnan(sxx).any() or np.isinf(sxx).any():
        sxx = np.nan_to_num(sxx)

    if len(sxx.shape) == 1:
        sum_c = np.mean(sxx, axis=0)
    elif len(sxx.shape) == 2:
        sum_c = np.mean(sxx, axis=1)
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(sxx.shape)))

    return sum_c


def just_tile_d0(d0_array1d_in: Union[float, np.ndarray],
                 d0d1_shape: tuple) -> np.ndarray:
    """
    Constructs tiled array from 1D array to the shape specified by d0d1_shape
    :param d0_array1d_in: 1D scalar or vector with dimension d0
    :param d0d1_shape: Tuple with output array shape, first dimension must match d0
    :return: ndarray
    """

    # TODO: Revisit - too many silly options
    if len(d0d1_shape) == 1:
        # Scalar to array, special case
        tiled_matrix = np.tile(d0_array1d_in, (d0d1_shape[0]))
    elif len(d0d1_shape) == 2:
        if d0d1_shape[0] == len(d0_array1d_in):
            tiled_matrix = np.tile(d0_array1d_in, (d0d1_shape[1], 1)).T
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d0_array1d_in.shape)))

    return tiled_matrix


def just_tile_d1(d1_array1d_in: Union[float, np.ndarray],
                 d0d1_shape: tuple) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply time-dependent values to frequency-time matrices

    :param d1: 1D input vector, nominally row time multipliers
    :param d0d1: 2D array, second dimension should be that same as d1
    :return: array with matching values
    """

    if len(d0d1_shape) == 1:
        tiled_matrix = np.tile(d1_array1d_in, (d0d1_shape[0]))
    elif len(d0d1_shape) == 2:
        if d0d1_shape[1] == len(d1_array1d_in):
            tiled_matrix = np.tile(d1_array1d_in, (d0d1_shape[0], 1))
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d1_array1d_in.shape)))

    return tiled_matrix


def sum_tile(sxx: np.ndarray) -> np.ndarray:
    """
    Compute the sum of the columns in a 1D or 2D array and then re-tile to the original size

    :param sxx: input vector or matrix
    :return: ndarray of sum
    """
    sum_c = sum_columns(sxx)

    # create array of repeated values of PSD with dimensions that match those of energy array
    if len(sxx.shape) == 1:
        sum_c_matrix = np.tile(sum_c, (sxx.shape[0]))
    elif len(sxx.shape) == 2:
        sum_c_matrix = np.tile(sum_c, (sxx.shape[1], 1)).T
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(sxx.shape)))

    return sum_c_matrix


def mean_tile(sxx: np.ndarray,
              shape_out: np.ndarray) -> np.ndarray:
    """
    Compute the mean of the columns in a 1D or 2D array and then re-tile to the original size

    :param sxx: input vector or matrix
    :param shape_out: shape of output vector or matrix
    :return: ndarray of mean
    """
    sum_c = mean_columns(sxx)

    # create array of repeated values of PSD with dimensions that match those of energy array
    if len(shape_out) == 1:
        sum_c_matrix = np.tile(sum_c, (shape_out[0]))
    elif len(shape_out) == 2:
        sum_c_matrix = np.tile(sum_c, (shape_out[1], 1)).T
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(sxx.shape)))

    return sum_c_matrix


def d0tile_x_d0d1(d0: Union[float, np.ndarray],
                  d0d1: np.ndarray) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply frequency-dependent values to frequency-time matrices

    :param d0: 1D input vector, nominally column frequency/scale multipliers
    :param d0d1: 2D array, first dimension should be that same as d1
    :return: array with matching values
    """
    # TODO: Add error catch, and firm up use case
    shape_out = d0d1.shape

    if len(shape_out) == 1:
        d0_matrix = np.tile(d0, (shape_out[0]))
    elif len(shape_out) == 2:
        d0_matrix = np.tile(d0, (shape_out[1], 1)).T
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d0.shape)))

    if d0_matrix.shape == d0d1.shape:
        d0_x_d0d1 = d0_matrix * d0d1
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d0.shape)))
    return d0_x_d0d1


def d1tile_x_d0d1(d1: Union[float, np.ndarray],
                  d0d1: np.ndarray) -> np.ndarray:
    """
    Create array of repeated values with dimensions that match those of energy array
    Useful to multiply time-dependent values to frequency-time matrices

    :param d1: 1D input vector, nominally row time multipliers
    :param d0d1: 2D array, second dimension should be that same as d1
    :return: array with matching values
    """
    shape_out = d0d1.shape

    if len(shape_out) == 1:
        d1_matrix = np.tile(d1, (shape_out[0]))
    elif len(shape_out) == 2:
        # TODO: Test
        d1_matrix = np.tile(d1, (shape_out[0], 1))
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d1.shape)))

    if d1_matrix.shape == d0d1.shape:
        d1_x_d0d1 = d1_matrix * d0d1
    else:
        raise TypeError('Cannot handle an array of shape {}.'.format(str(d1.shape)))
    return d1_x_d0d1


def decimate_array(sig_wf: np.array,
                   downsampling_factor: int) -> np.ndarray:
    """
    Decimate data and timestamps for an individual station
    All signals MUST have the same sample rate
    :param sig_wf: signal waveform
    :param downsampling_factor: the downsampling factor
    :param filter_order: the order of the filter
    :return: np.array decimated data
    """
    # decimate signal data
    decimated_data = signal.decimate(x=sig_wf,
                                    q=downsampling_factor,
                                    axis=1,
                                    zero_phase=True)

    return decimated_data



"""
This module contains plotting functions for Time Frequency Representations
"""
import enum
from typing import List, Tuple
import math
import numpy as np
from matplotlib.collections import QuadMesh
from matplotlib.colorbar import Colorbar
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable, AxesDivider
from dataclasses import dataclass
import quantum_inferno.utils_date_time as dt


# TODO: Add native color stylings
#  https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html
# TODO: For dark background
#  plt.style.use('dark_background')

class FigureAttributes:
    """
    This is the most basic parent class -- sets the plot canvas as well as figure handling functions. This is where
    figure size, font style and sizes, line weights and colors, etc. are established. All subsequent plot classes
    will inherit these attributes, overriding them if necessary.

    Attributes
    __________

    fig_size_ratio: 2d array of figure width, height ratio
    fontsize1_scale: int, scale for fontsize level 1 (titles, axes labels...)
    fontsize2: int, scale for fontsize level 2 (legend labels, ticks...)
    line_color: string, color for line in plot
    line_style: string, style of line in plot
    """

    def __init__(self, fig_size_ratio=np.array([640, 400]), fontsize1_scale=5, fontsize2_scale=4, line_color='k',
                 line_style='-'):

        self.fig_scale = 2.0
        self.fig_dpi = 300
        self.ratio = fig_size_ratio
        self.font_size_1st_level = np.rint(self.fig_scale * fontsize1_scale)
        self.font_size_2nd_level = np.rint(self.fig_scale * fontsize2_scale)
        self.line_color = line_color
        self.line_style = line_style
        self.fig_aspect_ratio = np.rint(self.fig_scale * self.ratio)  # was 640, 360, ratio 16:9
        self.fig_face_color = "w"
        self.fig_edge_color = self.fig_face_color
        self.fig_size = self.fig_aspect_ratio / self.fig_dpi
        self.font_color = "k"
        self.font_weight = "normal"
        self.line_weight = np.rint(self.fig_scale * 1)
        self.tick_size = self.font_size_2nd_level
        self.legend_label_size = self.font_size_2nd_level

        self.fig = None


class AspectRatioType(enum.Enum):
    """
    Enumeration denoting aspect ratios
    """
    R640x360 = 1
    R1280x720 = 2
    R1920x1080 = 3
    R2560x1440 = 4
    R3840x2160 = 5


class FigureParameters:
    """
    This class encapsulates the logic of computing figure size
    """

    def __init__(self, aspect_ratio: AspectRatioType):
        if aspect_ratio == AspectRatioType.R640x360:
            self.width = 640
            self.height = 360
            self.scale_factor = 1.0 / 3.0
        elif aspect_ratio == AspectRatioType.R1280x720:
            self.width = 1280
            self.height = 720
            self.scale_factor = 2.0 / 3.0
        elif aspect_ratio == AspectRatioType.R1920x1080:
            self.width = 1920
            self.height = 1080
            self.scale_factor = 1.25
        elif aspect_ratio == AspectRatioType.R2560x1440:
            self.width = 2560
            self.height = 1440
            self.scale_factor = 4.0 / 3.0
        else:
            self.width = 3840
            self.height = 2160
            self.scale_factor = 2.0

        scale = self.scale_factor * self.height / 8
        self.figure_size_x = int(self.width / scale)
        self.figure_size_y = int(self.height / scale)
        self.text_size = int(2.0 * self.height / scale)


# Set Aspect Ratio
@dataclass
class AudioParams:
    fill_gaps: bool = True
    figure_parameters: FigureParameters = FigureParameters(AspectRatioType.R1920x1080)


def origin_time_correction(time_input: np.ndarray,
                           start_time_epoch: float,
                           units_time: str,
                           start_time_sanitized: bool = True) -> Tuple[str, np.ndarray]:
    """
    Sanitize time
    :param time_input: array with timestamps
    :param start_time_epoch: start time in epoch UTC
    :param units_time: units of time
    :param start_time_sanitized: True or False
    :return: time label and time elapsed from start
    """
    # Sanitizing/scrubbing time is a key function.
    # TEST EXTENSIVELY!
    # Elapsed time from start of the record. Watch out with start alignment.
    # TODO: THIS WILL BREAK: TEST EXTENSIVELY!
    if start_time_sanitized:
        time_elapsed_from_start = time_input - time_input[0]
    else:
        time_elapsed_from_start = time_input
    # Need to construct a function to return time from start_time_epoch if they are not the same
    if start_time_epoch != time_input[0]:
        time_from_epoch_start = time_input[0] - start_time_epoch
    if start_time_epoch == 0:
        # Time sanitized if no input provided
        time_label: str = f"Time ({units_time})"
    else:
        start_datetime_epoch = dt.datetime.utcfromtimestamp(start_time_epoch)
        dt_str: str = start_datetime_epoch.strftime("%Y-%m-%d %H:%M:%S")
        time_label: str = f"Time ({units_time}) from UTC {dt_str}"

    return time_label, time_elapsed_from_start


def mesh_time_frequency_edges(frequency: np.ndarray,
                              time: np.ndarray,
                              frequency_ymin: float,
                              frequency_ymax: float,
                              frequency_scaling: str = "linear") -> Tuple[np.ndarray, np.ndarray, float, float]:
    """
    Find time and frequency edges for plotting

    :param frequency: array with frequencies
    :param time: array with timestamps
    :param frequency_ymin: minimum frequency for y-axis
    :param frequency_ymax: maximum frequency for y-axis
    :param frequency_scaling: "log" or "linear". Default is "linear"
    :return: min and max frequency for plot, time and frequency edges
    """

    if frequency_ymin > frequency_ymax:
        print("Higher frequency must be greater than lower frequency")
    if frequency[2] < frequency[1]:
        print("Frequency must be increasing, flip it")
    if time[2] < time[1]:
        print("Time must be increasing, flip it")

    t_half_bin: float = np.abs(time[2]-time[1]) / 2.
    t_edge: np.ndarray = np.append(time[0]-t_half_bin, time + t_half_bin)

    if frequency_scaling == "log":
        k_edge: float = np.sqrt(frequency[-1]/frequency[-2])
        f_edge: np.ndarray = np.append(frequency/k_edge, k_edge*frequency[-1])
    else:
        f_half_bin: float = (frequency[2]-frequency[1]) / 2.
        f_edge: np.ndarray = np.append(frequency[0]-f_half_bin, frequency + f_half_bin)

    # Initialize
    frequency_fix_ymin = 1.*frequency_ymin
    frequency_fix_ymax = 1.*frequency_ymax

    if frequency_ymin < f_edge[1]:
        frequency_fix_ymin = f_edge[0]
    if frequency_fix_ymin <= 0 and frequency_scaling == "log":
        frequency_fix_ymin = f_edge[1]
    if frequency_ymax > f_edge[-1]:
        frequency_fix_ymax = f_edge[-1]

    return t_edge, f_edge, frequency_fix_ymin, frequency_fix_ymax


def mesh_colormap_limits(mesh_array: np.ndarray,
                         colormap_scaling: str = "auto",
                         color_range: float = 16):
    """
    Find colormap limits for plotting

    :param mesh_array: array with mesh
    :param colormap_scaling: one of: "auto" (max/min of input mesh), "range" (max of input mesh minus color range given)
        or "abs" (absolute max/min of input mesh)
    :param color_range: default is 16.0
    :return: colormap min and max values
    """

    if colormap_scaling == "auto":
        color_max = np.max(mesh_array)
        color_min = np.min(mesh_array)
    elif colormap_scaling == "range":
        color_max = np.max(mesh_array)
        color_min = color_max - color_range
    else:
        print("Specify mesh color limits, using min max")
        color_max = np.max(np.abs(mesh_array))
        color_min = np.min(np.abs(mesh_array))

    return color_min, color_max


# # BEGIN PLOTS: THESE ARE THE MAIN TEMPLATES
def plot_wf_wf_wf_vert(station_id: str,
                       wf_panel_a_sig: np.ndarray,
                       wf_panel_a_time: np.ndarray,
                       wf_panel_b_sig: np.ndarray,
                       wf_panel_b_time: np.ndarray,
                       wf_panel_c_sig: np.ndarray,
                       wf_panel_c_time: np.ndarray,
                       start_time_epoch: float = 0,
                       start_time_sanitized: bool = True,
                       wf_panel_a_units: str = "Norm",
                       wf_panel_b_units: str = "Norm",
                       wf_panel_c_units: str = "Norm",
                       params_tfr=AudioParams(),
                       units_time: str = "s",
                       figure_title: str = "Time Domain Representation",
                       figure_title_show: bool = True,
                       label_panel_show: bool = False,
                       labels_panel_a: str = "(a)",
                       labels_panel_b: str = "(b)",
                       labels_panel_c: str = "(c)",
                       labels_fontweight: str = None) -> None:
    """
    Template for aligned time-series display

    :param station_id: name of station
    :param wf_panel_c_sig: array with signal waveform for top panel
    :param wf_panel_c_time: array with signal timestamps for top panel
    :param wf_panel_b_sig: array with signal waveform for middle panel
    :param wf_panel_b_time: array with signal timestamps for middle panel
    :param wf_panel_a_sig: array with signal waveform for bottom panel
    :param wf_panel_a_time: array with signal timestamps for bottom panel
    :param start_time_epoch: start time in epoch UTC. Default is 0.0
    :param wf_panel_c_units: units of signal (top panel). Default is "Norm"
    :param wf_panel_b_units: units of signal (middle panel). Default is "Norm"
    :param wf_panel_a_units: units of signal (bottom panel). Default is "Norm"
    :param params_tfr: parameters for tfr. Check AudioParams().
    :param waveform_color: color of waveforms. Default is "midnightblue"
    :param units_time: units of time. Default is "s"
    :param figure_title: title of figure. Default is "Time Domain Representation"
    :param figure_title_show: True to display title, False for publications
    :param label_panel_show: show panel labelling. Default is False, True for publication
    :param labels_panel_a: label for bottom panel. Default is (a)
    :param labels_panel_b: label for middle panel. Default is (b)
    :param labels_panel_c: label for top panel. Default is (c)
    :param labels_fontweight: matplotlib.text property
    :return: plot
    """

    if start_time_epoch == 0:
        time_label: str = f"Time ({units_time})"
        if start_time_sanitized:
            # Time sanitized if no input provided
            wf_panel_c_time_zero = wf_panel_c_time - wf_panel_a_time[0]
            wf_panel_b_time_zero = wf_panel_b_time - wf_panel_a_time[0]
            wf_panel_a_time_zero = wf_panel_a_time - wf_panel_a_time[0]
        else:
            wf_panel_c_time_zero = wf_panel_c_time
            wf_panel_b_time_zero = wf_panel_b_time
            wf_panel_a_time_zero = wf_panel_a_time
    else:
        # TODO: Allow for different time formats
        start_datetime_epoch = dt.datetime.utcfromtimestamp(start_time_epoch)
        dt_str: str = start_datetime_epoch.strftime("%Y-%m-%d %H:%M:%S")
        time_label: str = f"Time ({units_time}) from UTC {dt_str}"
        wf_panel_c_time_zero = wf_panel_c_time - start_time_epoch
        wf_panel_b_time_zero = wf_panel_b_time - start_time_epoch
        wf_panel_a_time_zero = wf_panel_a_time - start_time_epoch

    # Catch cases where there may not be any data
    time_xmin = wf_panel_a_time_zero[0]
    time_xmax = wf_panel_a_time_zero[-1]
    if time_xmin == time_xmax:
        time_xmin = wf_panel_b_time_zero[0]
        time_xmax = wf_panel_b_time_zero[-1]
    if time_xmin == time_xmax:
        time_xmin = wf_panel_c_time_zero[0]
        time_xmax = wf_panel_c_time_zero[-1]
    if time_xmin == time_xmax:
        print("No data to plot for " + figure_title)
        return

    # print("Panel time:", wf_panel_c_time_zero[0], wf_panel_c_time_zero[-1])
    # Figure starts here
    fig_ax_tuple: Tuple[plt.Figure, List[plt.Axes]] = \
        plt.subplots(3, 1,
                     figsize=(params_tfr.figure_parameters.figure_size_x,
                              params_tfr.figure_parameters.figure_size_y),
                     sharex=True)
    fig: plt.Figure = fig_ax_tuple[0]
    axes: List[plt.Axes] = fig_ax_tuple[1]
    wf_panel_c: plt.Axes = axes[0]
    wf_panel_b: plt.Axes = axes[1]
    wf_panel_a: plt.Axes = axes[2]

    if figure_title_show:
        wf_panel_c.set_title(f"{figure_title} at Station {station_id}")

    wf_panel_c.plot(wf_panel_c_time_zero, wf_panel_c_sig)
    if label_panel_show:
        wf_panel_c.text(0.01, 0.95, labels_panel_c, transform=wf_panel_c.transAxes,
                        fontsize=params_tfr.figure_parameters.text_size, fontweight=labels_fontweight, va='top')
    wf_panel_c.set_ylabel(wf_panel_c_units, size=params_tfr.figure_parameters.text_size)
    wf_panel_c.set_xlim(time_xmin, time_xmax)
    wf_panel_c.tick_params(axis='x', which='both', bottom=False, labelbottom=False, labelsize='large')
    wf_panel_c.grid(True)
    wf_panel_c.tick_params(axis='y', labelsize='large')
    wf_panel_c.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")
    wf_panel_c.yaxis.get_offset_text().set_x(-0.034)

    wf_panel_b.plot(wf_panel_b_time_zero, wf_panel_b_sig)
    if label_panel_show:
        wf_panel_b.text(0.01, 0.95, labels_panel_b, transform=wf_panel_b.transAxes,
                        fontsize=params_tfr.figure_parameters.text_size, fontweight=labels_fontweight, va='top')
    wf_panel_b.set_ylabel(wf_panel_b_units, size=params_tfr.figure_parameters.text_size)
    wf_panel_b.set_xlim(time_xmin, time_xmax)
    wf_panel_b.tick_params(axis='x', which='both', bottom=False, labelbottom=False, labelsize='large')
    wf_panel_b.grid(True)
    wf_panel_b.tick_params(axis='y', labelsize='large')
    wf_panel_b.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")
    wf_panel_b.yaxis.get_offset_text().set_x(-0.034)

    wf_panel_a.plot(wf_panel_a_time_zero, wf_panel_a_sig)
    if label_panel_show:
        wf_panel_a.text(0.01, 0.95, labels_panel_a, transform=wf_panel_a.transAxes,
                        fontsize=params_tfr.figure_parameters.text_size, fontweight=labels_fontweight, va='top')
    wf_panel_a.set_ylabel(wf_panel_a_units, size=params_tfr.figure_parameters.text_size)
    wf_panel_a.set_xlim(time_xmin, time_xmax)
    wf_panel_a.tick_params(axis='x', which='both', bottom=False, labelbottom=True, labelsize='large')
    wf_panel_a.grid(True)
    wf_panel_a.tick_params(axis='y', labelsize='large')
    wf_panel_a.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")
    wf_panel_a.yaxis.get_offset_text().set_x(-0.034)

    fig.text(.5, .01, time_label, ha='center', size=params_tfr.figure_parameters.text_size)

    fig.align_ylabels(axes)
    fig.tight_layout()
    fig.subplots_adjust(bottom=.1, hspace=0.13)


def plot_wf_mesh_mesh_vert(station_id: str,
                           wf_panel_a_sig: np.ndarray,
                           wf_panel_a_time: np.ndarray,
                           mesh_time: np.ndarray,
                           mesh_frequency: np.ndarray,
                           mesh_panel_b_tfr: np.ndarray,
                           mesh_panel_c_tfr: np.ndarray,
                           params_tfr=AudioParams(),
                           frequency_scaling: str = "log",
                           mesh_shading: str = "auto",
                           mesh_panel_b_colormap_scaling: str = "auto",
                           mesh_panel_b_color_max: float = 15,
                           mesh_panel_b_color_range: float = 15,
                           mesh_panel_b_color_min: float = 0,
                           mesh_panel_c_colormap_scaling: str = "auto",
                           mesh_panel_c_color_max: float = 15,
                           mesh_panel_c_color_range: float = 15,
                           mesh_panel_c_color_min: float = 0,
                           start_time_epoch: float = 0,
                           frequency_hz_ymin: float = None,
                           frequency_hz_ymax: float = None,
                           mesh_colormap: str = "inferno",
                           units_time: str = "s",
                           units_frequency: str = "Hz",
                           wf_panel_a_units: str = "Norm",
                           mesh_panel_b_cbar_units: str = "bits",
                           mesh_panel_c_cbar_units: str = "bits",
                           figure_title: str = "Time-Frequency Representation",
                           figure_title_show: bool = True):

    """
    Plot 3 vertical panels - mesh (top panel), mesh (middle panel) and signal waveform (bottom panel)

    :param station_id: name of station
    :param wf_panel_a_sig: array with signal waveform for bottom panel
    :param wf_panel_a_time: array with signal timestamps for bottom panel
    :param mesh_time: array with mesh time
    :param mesh_frequency: array with mesh frequencies
    :param mesh_panel_b_tfr: array with mesh tfr data for mesh plot (middle panel)
    :param mesh_panel_c_tfr: array with mesh tfr data for mesh plot (top panel)
    :param params_tfr: parameters for tfr. Check AudioParams().
    :param frequency_scaling: "log" or "linear". Default is "log"
    :param mesh_shading: type of mesh shading, one of "auto", "gouraud" or "else". Default is "auto"
     :param mesh_panel_b_colormap_scaling: color scaling for mesh plot (middle panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_b_color_max, mesh_panel_b_color_range, mesh_panel_b_color_min). Default is "auto"
    :param mesh_panel_b_color_max: maximum value for color scaling for mesh plot (middle panel). Default is 15.0
    :param mesh_panel_b_color_range: range between maximum and minimum values in color scaling for mesh plot
        (middle panel). Default is 15.0
    :param mesh_panel_b_color_min: minimum value for color scaling for mesh plot (middle panel). Default is 0.0
    :param mesh_panel_c_colormap_scaling: color scaling for mesh plot (top panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_c_color_max, mesh_panel_c_color_range, mesh_panel_c_color_min). Default is "auto"
    :param mesh_panel_c_color_max: maximum value for color scaling for mesh plot (top panel). Default is 15.0
    :param mesh_panel_c_color_range:range between maximum and minimum values in color scaling for scatter plot
        (top panel). Default is 15.0
    :param mesh_panel_c_color_min: minimum value for color scaling for mesh plot (top panel). Default is 0.0
    :param start_time_epoch: start time in epoch UTC. Default is 0.0
    :param frequency_hz_ymin: minimum frequency for y-axis
    :param frequency_hz_ymax: maximum frequency for y-axis
    :param waveform_color: color of waveform for bottom panel. Default is "midnightblue"
    :param mesh_colormap: a Matplotlib Colormap instance or registered colormap name. Default is "inferno"
    :param units_time: units of time. Default is "s"
    :param units_frequency: units of frequency. Default is "Hz"
    :param wf_panel_a_units: units of waveform plot (bottom panel). Default is "Norm"
    :param mesh_panel_b_cbar_units: units of colorbar for mesh plot (middle panel). Default is "bits"
    :param mesh_panel_c_cbar_units: units of colorbar for mesh plot (top panel). Default is "bits"
    :param figure_title: title of figure. Default is "Time-Frequency Representation"
    :param figure_title_show: show title if True. Default is True
    :return: plot
    """

    # Autoscale to mesh frequency range
    if frequency_hz_ymax is None:
        frequency_hz_ymax = mesh_frequency[-1]
    if frequency_hz_ymin is None:
        frequency_hz_ymin = mesh_frequency[0]

    # Time zeroing and scrubbing, if needed
    time_label, wf_panel_a_elapsed_time = \
        origin_time_correction(wf_panel_a_time, start_time_epoch, units_time)

    # Time is in the center of the window, frequency is in the fft coefficient center.
    # pcolormesh must provide corner coordinates, so there will be an offset from step noverlap step size.
    # frequency and time must be increasing!
    t_edge, f_edge, frequency_fix_ymin, frequency_fix_ymax = \
        mesh_time_frequency_edges(frequency=mesh_frequency, time=mesh_time,
                                  frequency_ymin=frequency_hz_ymin,
                                  frequency_ymax=frequency_hz_ymax,
                                  frequency_scaling=frequency_scaling)

    # Figure starts here
    fig_ax_tuple: Tuple[plt.Figure, List[plt.Axes]] = \
        plt.subplots(3, 1,
                     figsize=(params_tfr.figure_parameters.figure_size_x,
                              params_tfr.figure_parameters.figure_size_y),
                     sharex=True)
    fig: plt.Figure = fig_ax_tuple[0]
    axes: List[plt.Axes] = fig_ax_tuple[1]
    mesh_panel_c: plt.Axes = axes[0]
    mesh_panel_b: plt.Axes = axes[1]
    wf_panel_a: plt.Axes = axes[2]
    # bottom_panel_picker: plt.Axes = axes[3]

    # Top panel mesh --------------------------
    # Time is in the center of the window, frequency is in the fft coefficient center.
    # pcolormesh must provide corner coordinates, so there will be an offset from step noverlap step size.
    # frequency and time must be increasing!

    # Display preference
    wf_panel_a_time_xmin: int = wf_panel_a_elapsed_time[0]
    wf_panel_a_time_xmax: int = t_edge[-1]

    # Override, default is autoscaling to min and max values
    if mesh_panel_b_colormap_scaling == "auto":
        mesh_panel_b_color_min, mesh_panel_b_color_max = mesh_colormap_limits(mesh_panel_b_tfr,
                                                                              mesh_panel_b_colormap_scaling,
                                                                              mesh_panel_b_color_range)
    elif mesh_panel_b_colormap_scaling == "range":
        mesh_panel_b_color_min, mesh_panel_b_color_max = mesh_colormap_limits(mesh_panel_b_tfr,
                                                                              mesh_panel_b_colormap_scaling,
                                                                              mesh_panel_b_color_range)
    else:
        "Mesh 1 color scaling with user inputs"

    if mesh_panel_c_colormap_scaling == "auto":
        mesh_panel_c_color_min, mesh_panel_c_color_max = mesh_colormap_limits(mesh_panel_c_tfr,
                                                                              mesh_panel_c_colormap_scaling,
                                                                              mesh_panel_c_color_range)
    elif mesh_panel_c_colormap_scaling == "range":
        mesh_panel_c_color_min, mesh_panel_c_color_max = mesh_colormap_limits(mesh_panel_c_tfr,
                                                                              mesh_panel_c_colormap_scaling,
                                                                              mesh_panel_c_color_range)
    else:
        "Mesh 0 color scaling with user inputs"

    # Setup color map ticks
    all_cbar_ticks_lens: List[int] = [
        len(str(math.ceil(mesh_panel_c_color_min))),
        len(str(math.floor(mesh_panel_c_color_max))),
        len(str(math.ceil(mesh_panel_b_color_min))),
        len(str(math.floor(mesh_panel_b_color_max)))
    ]
    max_cbar_tick_len: int = sorted(all_cbar_ticks_lens)[-1]
    cbar_tick_fmt: str = f"%-{max_cbar_tick_len}s"

    if mesh_shading == "auto":
        pcolormesh_top: QuadMesh = mesh_panel_c.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_c_tfr,
                                                           vmin=mesh_panel_c_color_min,
                                                           vmax=mesh_panel_c_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    elif mesh_shading == "gouraud":
        pcolormesh_top: QuadMesh = mesh_panel_c.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_c_tfr,
                                                           vmin=mesh_panel_c_color_min,
                                                           vmax=mesh_panel_c_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    else:
        pcolormesh_top: QuadMesh = mesh_panel_c.pcolormesh(t_edge,
                                                           f_edge,
                                                           mesh_panel_c_tfr,
                                                           vmin=mesh_panel_c_color_min,
                                                           vmax=mesh_panel_c_color_max,
                                                           cmap=mesh_colormap,
                                                           snap=True)

    mesh_panel_c_div: AxesDivider = make_axes_locatable(mesh_panel_c)
    mesh_panel_c_cax: plt.Axes = mesh_panel_c_div.append_axes("right", size="1%", pad="0.5%")
    mesh_panel_c_cbar: Colorbar = fig.colorbar(pcolormesh_top, cax=mesh_panel_c_cax,
                                               ticks=[math.ceil(mesh_panel_c_color_min),
                                                      math.floor(mesh_panel_c_color_max)],
                                               format=cbar_tick_fmt)
    mesh_panel_c_cbar.set_label(mesh_panel_c_cbar_units, rotation=270, size=params_tfr.figure_parameters.text_size)
    mesh_panel_c_cax.tick_params(labelsize='large')
    # if figure_title_show:
    #     mesh_panel_c.set_title(f"{figure_title} at Station {station_id}")
    if figure_title_show:
        mesh_panel_c.set_title(f"{figure_title} ({station_id})")
    mesh_panel_c.set_ylabel(units_frequency, size=params_tfr.figure_parameters.text_size)
    mesh_panel_c.set_xlim(wf_panel_a_time_xmin, wf_panel_a_time_xmax)
    mesh_panel_c.set_ylim(frequency_fix_ymin, frequency_fix_ymax)
    mesh_panel_c.set_yscale(frequency_scaling)
    mesh_panel_c.tick_params(axis='x', which='both', bottom=False, labelbottom=False)
    mesh_panel_c.tick_params(axis='y', labelsize='large')

    # Middle panel mesh --------------------------

    if mesh_shading == "auto":
        pcolormesh_mid: QuadMesh = mesh_panel_b.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    elif mesh_shading == "gouraud":
        pcolormesh_mid: QuadMesh = mesh_panel_b.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    else:
        pcolormesh_mid: QuadMesh = mesh_panel_b.pcolormesh(t_edge,
                                                           f_edge,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           snap=True)

    mesh_panel_b_div: AxesDivider = make_axes_locatable(mesh_panel_b)
    mesh_panel_b_cax: plt.Axes = mesh_panel_b_div.append_axes("right", size="1%", pad="0.5%")
    mesh_panel_b_cbar: Colorbar = fig.colorbar(pcolormesh_mid, cax=mesh_panel_b_cax,
                                               ticks=[math.ceil(mesh_panel_b_color_min),
                                                      math.floor(mesh_panel_b_color_max)],
                                               format=cbar_tick_fmt)
    mesh_panel_b_cbar.set_label(mesh_panel_b_cbar_units, rotation=270, size=params_tfr.figure_parameters.text_size)
    mesh_panel_b_cax.tick_params(labelsize='large')

    mesh_panel_b.set_ylabel(units_frequency, size=params_tfr.figure_parameters.text_size)
    mesh_panel_b.set_xlim(wf_panel_a_time_xmin, wf_panel_a_time_xmax)
    mesh_panel_b.set_ylim(frequency_fix_ymin, frequency_fix_ymax)
    mesh_panel_b.margins(x=0)
    mesh_panel_b.set_yscale(frequency_scaling)
    mesh_panel_b.tick_params(axis='x', which='both', bottom=False, labelbottom=False)
    mesh_panel_b.tick_params(axis='y', labelsize='large')

    # Waveform panel
    wf_panel_a.plot(wf_panel_a_elapsed_time, wf_panel_a_sig)
    wf_panel_a.set_ylabel(wf_panel_a_units, size=params_tfr.figure_parameters.text_size)
    wf_panel_a.set_xlim(wf_panel_a_time_xmin, wf_panel_a_time_xmax)
    wf_panel_a.tick_params(axis='x', which='both', bottom=True, labelbottom=True, labelsize='large')
    wf_panel_a.grid(True)
    wf_panel_a.tick_params(axis='y', labelsize='large')

    # TODO: ADD OPTIONS AS BELOW
    # wf_panel_a.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")
    wf_panel_a.ticklabel_format(style="plain", scilimits=(0, 0), axis="y")
    wf_panel_a.yaxis.get_offset_text().set_x(-0.034)

    wf_panel_a_div: AxesDivider = make_axes_locatable(wf_panel_a)
    wf_panel_a_cax: plt.Axes = wf_panel_a_div.append_axes("right", size="1%", pad="0.5%")
    wf_panel_a_cax.axis("off")

    fig.text(.5, .01, time_label, ha='center', size=params_tfr.figure_parameters.text_size)

    fig.align_ylabels(axes)
    fig.tight_layout()
    fig.subplots_adjust(bottom=.1, hspace=0.13)

    return fig


def plot_wf_mesh_vert(station_id: str,
                      wf_panel_a_sig: np.ndarray,
                      wf_panel_a_time: np.ndarray,
                      mesh_time: np.ndarray,
                      mesh_frequency: np.ndarray,
                      mesh_panel_b_tfr: np.ndarray,
                      params_tfr=AudioParams(),
                      frequency_scaling: str = "log",
                      mesh_shading: str = "auto",
                      wf_panel_a_yscaling: str = "auto",
                      wf_panel_a_ytick_style: str = "plain",
                      mesh_panel_b_ytick_style: str = 'sci',
                      mesh_panel_b_colormap_scaling: str = "auto",
                      mesh_panel_b_color_max: float = 15,
                      mesh_panel_b_color_range: float = 15,
                      mesh_panel_b_color_min: float = 0,
                      start_time_epoch: float = 0,
                      start_time_sanitized: bool = True,
                      frequency_hz_ymin: float = None,
                      frequency_hz_ymax: float = None,
                      mesh_colormap: str = None,
                      waveform_color: str = None,
                      units_time: str = "s",
                      units_frequency: str = "Hz",
                      wf_panel_a_units: str = "Norm",
                      mesh_panel_b_cbar_units: str = "bits",
                      figure_title: str = "Time-Frequency Representation",
                      figure_title_show: bool = True):
    """
    Plot 2 vertical panels - mesh (top panel) and signal waveform (bottom panel)

    :param wf_panel_a_yscaling: 'auto', 'symmetric', 'positive'
    :param station_id: name of station
    :param wf_panel_a_sig: array with signal waveform for bottom panel
    :param wf_panel_a_time: array with signal timestamps for bottom panel
    :param mesh_time: array with mesh time
    :param mesh_frequency: array with mesh frequencies
    :param mesh_panel_b_tfr: array with mesh tfr data for mesh plot (top panel)
    :param params_tfr: parameters for tfr. Check AudioParams().
    :param frequency_scaling: "log" or "linear". Default is "log"
    :param mesh_shading: type of mesh shading, one of "auto", "gouraud" or "else". Default is "auto"
    :param mesh_panel_b_colormap_scaling: color scaling for mesh plot (top panel). One of: "auto", "range" or "else"
        (use inputs given in mesh_panel_b_color_max, mesh_panel_b_color_range, mesh_panel_b_color_min). Default is "auto"
    :param mesh_panel_b_color_max: maximum value for color scaling for mesh plot (top panel). Default is 15.0
    :param mesh_panel_b_color_range:range between maximum and minimum values in color scaling for scatter plot
        (top panel). Default is 15.0
    :param mesh_panel_b_color_min: minimum value for color scaling for mesh plot (top panel). Default is 0.0
    :param start_time_epoch: start time in epoch UTC. Default is 0.0
    :param frequency_hz_ymin: minimum frequency for y-axis
    :param frequency_hz_ymax: maximum frequency for y-axis
    :param waveform_color: color of waveform for bottom panel. Default is "midnightblue"
    :param mesh_colormap: a Matplotlib Colormap instance or registered colormap name. If None, inherits style sheet spec
    :param units_time: units of time. Default is "s"
    :param units_frequency: units of frequency. Default is "Hz"
    :param wf_panel_a_units: units of waveform plot (bottom panel). Default is "Norm"
    :param wf_panel_a_ytick_style: 'plain' or 'sci'. Default is "plain"
    :param mesh_panel_b_ytick_style: 'plain' or 'sci'. Default is "sci"
    :param mesh_panel_b_cbar_units: units of colorbar for mesh plot (top panel). Default is "bits"
    :param figure_title: title of figure. Default is "Time-Frequency Representation"
    :param figure_title_show: show title if True. Default is True
    :return: plot
    """

    # Autoscale to mesh frequency range
    if frequency_hz_ymax is None:
        frequency_hz_ymax = mesh_frequency[-1]
    if frequency_hz_ymin is None:
        frequency_hz_ymin = mesh_frequency[0]

    # Time zeroing and scrubbing, if needed
    time_label, wf_panel_a_elapsed_time = \
        origin_time_correction(wf_panel_a_time, start_time_epoch, units_time, start_time_sanitized)

    # Time is in the center of the window, frequency is in the fft coefficient center.
    # pcolormesh must provide corner coordinates, so there will be an offset from step noverlap step size.
    # frequency and time must be increasing!
    t_edge, f_edge, frequency_fix_ymin, frequency_fix_ymax = \
        mesh_time_frequency_edges(frequency=mesh_frequency, time=mesh_time,
                                  frequency_ymin=frequency_hz_ymin,
                                  frequency_ymax=frequency_hz_ymax,
                                  frequency_scaling=frequency_scaling)

    # Figure starts here
    fig_ax_tuple: Tuple[plt.Figure, List[plt.Axes]] = \
        plt.subplots(2, 1,
                     figsize=(params_tfr.figure_parameters.figure_size_x,
                              params_tfr.figure_parameters.figure_size_y),
                     sharex=True)
    fig: plt.Figure = fig_ax_tuple[0]
    axes: List[plt.Axes] = fig_ax_tuple[1]
    mesh_panel_b: plt.Axes = axes[0]
    wf_panel_a: plt.Axes = axes[1]
    # bottom_panel_picker: plt.Axes = axes[3]

    # Top panel mesh --------------------------
    # Time is in the center of the window, frequency is in the fft coefficient center.
    # pcolormesh must provide corner coordinates, so there will be an offset from step noverlap step size.
    # frequency and time must be increasing!

    # Display preference
    wf_panel_a_time_xmin: int = wf_panel_a_elapsed_time[0]
    wf_panel_a_time_xmax: int = t_edge[-1]

    if mesh_panel_b_colormap_scaling == "auto":
        mesh_panel_b_color_min, mesh_panel_b_color_max = mesh_colormap_limits(mesh_panel_b_tfr,
                                                                              mesh_panel_b_colormap_scaling,
                                                                              mesh_panel_b_color_range)
    elif mesh_panel_b_colormap_scaling == "range":
        mesh_panel_b_color_min, mesh_panel_b_color_max = mesh_colormap_limits(mesh_panel_b_tfr,
                                                                              mesh_panel_b_colormap_scaling,
                                                                              mesh_panel_b_color_range)
    else:
        "Mesh 0 color scaling with user inputs"

    # Setup color map ticks
    all_cbar_ticks_lens: List[int] = [
        len(str(math.ceil(mesh_panel_b_color_min))),
        len(str(math.floor(mesh_panel_b_color_max)))]
    max_cbar_tick_len: int = sorted(all_cbar_ticks_lens)[-1]
    cbar_tick_fmt: str = f"%-{max_cbar_tick_len}s"

    if mesh_shading == "auto":
        pcolormesh_top: QuadMesh = mesh_panel_b.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    elif mesh_shading == "gouraud":
        pcolormesh_top: QuadMesh = mesh_panel_b.pcolormesh(mesh_time,
                                                           mesh_frequency,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           shading=mesh_shading,
                                                           snap=True)
    else:
        pcolormesh_top: QuadMesh = mesh_panel_b.pcolormesh(t_edge,
                                                           f_edge,
                                                           mesh_panel_b_tfr,
                                                           vmin=mesh_panel_b_color_min,
                                                           vmax=mesh_panel_b_color_max,
                                                           cmap=mesh_colormap,
                                                           snap=True)

    mesh_panel_b_div: AxesDivider = make_axes_locatable(mesh_panel_b)
    mesh_panel_b_cax: plt.Axes = mesh_panel_b_div.append_axes("right", size="1%", pad="0.5%")
    mesh_panel_b_cbar: Colorbar = fig.colorbar(pcolormesh_top, cax=mesh_panel_b_cax,
                                               ticks=[math.ceil(mesh_panel_b_color_min),
                                                      math.floor(mesh_panel_b_color_max)],
                                               format=cbar_tick_fmt)
    # mesh_panel_b_cbar: Colorbar = fig.colorbar(pcolormesh, cax=mesh_panel_b_cax,
    #                                         ticks=[colormin_top, colormax_top],
    #                                         format=cbar_tick_fmt)
    mesh_panel_b_cbar.set_label(mesh_panel_b_cbar_units, rotation=270, size=params_tfr.figure_parameters.text_size)
    mesh_panel_b_cax.tick_params(labelsize='large')
    if figure_title_show:
        mesh_panel_b.set_title(f"{figure_title} at Station {station_id}")
    mesh_panel_b.set_ylabel(units_frequency, size=params_tfr.figure_parameters.text_size)
    mesh_panel_b.set_xlim(wf_panel_a_time_xmin, wf_panel_a_time_xmax)
    mesh_panel_b.set_ylim(frequency_fix_ymin, frequency_fix_ymax)
    # mesh_panel_b.get_xaxis().set_ticklabels([])
    mesh_panel_b.set_yscale(frequency_scaling)
    mesh_panel_b.tick_params(axis='x', which='both', bottom=False, labelbottom=False)
    mesh_panel_b.tick_params(axis='y', labelsize='large')
    if frequency_scaling == "linear":
        # Only works for linear range
        mesh_panel_b.ticklabel_format(style=mesh_panel_b_ytick_style, scilimits=(0, 0), axis="y")
    # Waveform panel
    wf_panel_a.plot(wf_panel_a_elapsed_time, wf_panel_a_sig, color=waveform_color)
    wf_panel_a.set_ylabel(wf_panel_a_units, size=params_tfr.figure_parameters.text_size)
    wf_panel_a.set_xlim(wf_panel_a_time_xmin, wf_panel_a_time_xmax)

    wf_panel_a.tick_params(axis='x', which='both', bottom=True, labelbottom=True, labelsize='large')
    wf_panel_a.grid(True)
    wf_panel_a.tick_params(axis='y', labelsize='large')
    # TODO: Standardize with inputs
    if wf_panel_a_yscaling == 'auto':
        wf_panel_a.set_ylim(np.min(wf_panel_a_sig), np.max(wf_panel_a_sig))
        wf_panel_a.ticklabel_format(style="plain", scilimits=(0, 0), axis="y")
    elif wf_panel_a_yscaling == 'symmetric':
        wf_panel_a.set_ylim(-np.max(np.abs(wf_panel_a_sig)), np.max(np.abs(wf_panel_a_sig)))
    elif wf_panel_a_yscaling == 'positive':
        wf_panel_a.set_ylim(0, np.max(np.abs(wf_panel_a_sig)))
    else:
        wf_panel_a.set_ylim(-10, 10)
    wf_panel_a.ticklabel_format(style=wf_panel_a_ytick_style, scilimits=(0, 0), axis="y")
    wf_panel_a.yaxis.get_offset_text().set_x(-0.034)

    wf_panel_a_div: AxesDivider = make_axes_locatable(wf_panel_a)
    wf_panel_a_cax: plt.Axes = wf_panel_a_div.append_axes("right", size="1%", pad="0.5%")
    wf_panel_a_cax.axis("off")

    fig.text(.5, .01, time_label, ha='center', size=params_tfr.figure_parameters.text_size)

    fig.align_ylabels(axes)
    fig.tight_layout()
    fig.subplots_adjust(bottom=.1, hspace=0.13)

    return fig


"""
This RedVox SDK module contains constants and helper functions for converting between different time bases.
All time-based functions take inputs and output in UTC.
"""

from datetime import datetime, tzinfo, timedelta
from typing import List, Tuple

# noinspection Mypy
import numpy as np

# Microsecond constants
MICROSECONDS_IN_MILLISECOND: float = 1000.0
MICROSECONDS_IN_SECOND: float = 1000000.0
MICROSECONDS_IN_MINUTE: float = 60000000.0
MICROSECONDS_IN_HOUR: float = 3600000000.0
MICROSECONDS_IN_DAY: float = 86400000000.0
MICROSECONDS_IN_WEEK: float = 604800000000.0

# Millisecond constants
MILLISECONDS_IN_SECOND: float = 1000.0
MILLISECONDS_IN_MINUTE: float = 60000.0
MILLISECONDS_IN_HOUR: float = 3600000.0
MILLISECONDS_IN_DAY: float = 86400000.0
MILLISECONDS_IN_WEEK: float = 604800000.0

# Second constants
SECONDS_IN_MINUTE: float = 60.0
SECONDS_IN_HOUR: float = 3600.0
SECONDS_IN_DAY: float = 86400.0
SECONDS_IN_WEEK: float = 604800.0

# Minute constants
MINUTES_IN_HOUR: float = 60.0
MINUTES_IN_DAY: float = 1440.0
MINUTES_IN_WEEK: float = 10080.0

# Hour constants
HOURS_IN_DAY: float = 24.0
HOURS_IN_WEEK: float = 168.0

# Day constants
DAYS_IN_WEEK: float = 7.0

# A datetime object representing the epoch.
EPOCH: datetime = datetime.utcfromtimestamp(0)


class UTC(tzinfo):
    """
    This class represents UTC time.
    """

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


# An instance of the above class that can be used anywhere the class is needed
UTC_INSTANCE = UTC()


# Microsecond conversions
def microseconds_to_milliseconds(microseconds: float) -> float:
    """
    Converts microseconds to milliseconds.

    :param microseconds: Number of microseconds.
    :return: Number of milliseconds.
    """
    return microseconds / MICROSECONDS_IN_MILLISECOND


def microseconds_to_seconds(microseconds: float) -> float:
    """
    Converts microseconds to seconds.

    :param microseconds: Number of microseconds.
    :return: Number of seconds.
    """
    return microseconds / MICROSECONDS_IN_SECOND


def microseconds_to_minutes(microseconds: float) -> float:
    """
    Converts microseconds to minutes.

    :param microseconds: Number of microseconds.
    :return: Number of minutes.
    """
    return microseconds / MICROSECONDS_IN_MINUTE


def microseconds_to_hours(microseconds: float) -> float:
    """
    Converts microseconds to hours.

    :param microseconds: Number of microseconds.
    :return: Number of hours.
    """
    return microseconds / MICROSECONDS_IN_HOUR


def microseconds_to_days(microseconds: float) -> float:
    """
    Converts microseconds to days.

    :param microseconds: Number of microseconds.
    :return: Number of days.
    """
    return microseconds / MICROSECONDS_IN_DAY


def microseconds_to_weeks(microseconds: float) -> float:
    """
    Converts microseconds to weeks.

    :param microseconds: Number of microseconds.
    :return: Number of weeks.
    """
    return microseconds / MICROSECONDS_IN_WEEK


# Millisecond conversions
def milliseconds_to_microseconds(milliseconds: float) -> float:
    """
    Converts milliseconds to microseconds.

    :param milliseconds: Number of milliseconds.
    :return: Number of microseconds.
    """
    return milliseconds * MICROSECONDS_IN_MILLISECOND


def milliseconds_to_seconds(milliseconds: float) -> float:
    """
    Converts milliseconds to seconds.

    :param milliseconds: Number of milliseconds.
    :return: Number of seconds.
    """
    return milliseconds / MILLISECONDS_IN_SECOND


def milliseconds_to_minutes(milliseconds: float) -> float:
    """
    Converts milliseconds to minutes.

    :param milliseconds: Number of milliseconds.
    :return: Number of minutes.
    """
    return milliseconds / MILLISECONDS_IN_MINUTE


def milliseconds_to_hours(milliseconds: float) -> float:
    """
    Converts milliseconds to hours.

    :param milliseconds: Number of milliseconds.
    :return: Number of hours.
    """
    return milliseconds / MILLISECONDS_IN_HOUR


def milliseconds_to_days(milliseconds: float) -> float:
    """
    Converts milliseconds to days.

    :param milliseconds: Number of milliseconds.
    :return: Number of days.
    """
    return milliseconds / MILLISECONDS_IN_DAY


def milliseconds_to_weeks(milliseconds: float) -> float:
    """
    Converts milliseconds to weeks.

    :param milliseconds: Number of milliseconds.
    :return: Number of weeks.
    """
    return milliseconds / MILLISECONDS_IN_WEEK


# Second conversions
def seconds_to_microseconds(seconds: float) -> float:
    """
    Converts seconds to microseconds.

    :param seconds: Number of seconds.
    :return: Number of microseconds.
    """
    return seconds * MICROSECONDS_IN_SECOND


def seconds_to_milliseconds(seconds: float) -> float:
    """
    Converts seconds to milliseconds.

    :param seconds: Number of seconds.
    :return: Number of milliseconds.
    """
    return seconds * MILLISECONDS_IN_SECOND


def seconds_to_minutes(seconds: float) -> float:
    """
    Converts seconds to minutes.

    :param seconds: Number of seconds.
    :return: Number of minutes.
    """
    return seconds / SECONDS_IN_MINUTE


def seconds_to_hours(seconds: float) -> float:
    """
    Converts seconds to hours.

    :param seconds: Number of seconds.
    :return: Number of hours.
    """
    return seconds / SECONDS_IN_HOUR


def seconds_to_days(seconds: float) -> float:
    """
    Converts seconds to days.

    :param seconds: Number of seconds.
    :return: Number of days.
    """
    return seconds / SECONDS_IN_DAY


def seconds_to_weeks(seconds: float) -> float:
    """
    Converts seconds to weeks.

    :param seconds: Number of seconds.
    :return: Number of weeks.
    """
    return seconds / SECONDS_IN_WEEK


# Minute conversions
def minutes_to_microseconds(minutes: float) -> float:
    """
    Converts minutes to microseconds.

    :param minutes: Number of minutes.
    :return: Number of microseconds.
    """
    return minutes * MICROSECONDS_IN_MINUTE


def minutes_to_milliseconds(minutes: float) -> float:
    """
    Converts minutes to milliseconds.

    :param minutes: Number of minutes.
    :return: Number of milliseconds.
    """
    return minutes * MILLISECONDS_IN_MINUTE


def minutes_to_seconds(minutes: float) -> float:
    """
    Converts minutes to seconds.

    :param minutes: Number of minutes.
    :return: Number of seconds.
    """
    return minutes * SECONDS_IN_MINUTE


def minutes_to_hours(minutes: float) -> float:
    """
    Converts minutes to hours.

    :param minutes: Number of minutes.
    :return: Number of hours.
    """
    return minutes / MINUTES_IN_HOUR


def minutes_to_days(minutes: float) -> float:
    """
    Converts minutes to days.

    :param minutes: Number of minutes.
    :return: Number of days.
    """
    return minutes / MINUTES_IN_DAY


def minutes_to_weeks(minutes: float) -> float:
    """
    Converts minutes to weeks.

    :param minutes: Number of minutes.
    :return: Number of weeks.
    """
    return minutes / MINUTES_IN_WEEK


# Hour conversions
def hours_to_microseconds(hours: float) -> float:
    """
    Converts hours to microseconds.

    :param hours: Number of hours.
    :return: Number of microseconds.
    """
    return hours * MICROSECONDS_IN_HOUR


def hours_to_milliseconds(hours: float) -> float:
    """
    Converts hours to milliseconds.

    :param hours: Number of hours.
    :return: Number of milliseconds.
    """
    return hours * MILLISECONDS_IN_HOUR


def hours_to_seconds(hours: float) -> float:
    """
    Converts hours to seconds.

    :param hours: Number of hours.
    :return: Number of seconds.
    """
    return hours * SECONDS_IN_HOUR


def hours_to_minutes(hours: float) -> float:
    """
    Converts hours to minutes.

    :param hours: Number of hours.
    :return: Number of minutes.
    """
    return hours * MINUTES_IN_HOUR


def hours_to_days(hours: float) -> float:
    """
    Converts hours to days.

    :param hours: Number of hours.
    :return: Number of days.
    """
    return hours / HOURS_IN_DAY


def hours_to_weeks(hours: float) -> float:
    """
    Converts hours to weeks.

    :param hours: Number of hours.
    :return: Number of weeks.
    """
    return hours / HOURS_IN_WEEK


# Week conversions
def weeks_to_microseconds(weeks: float) -> float:
    """
    Converts weeks to microseconds.

    :param weeks: Number of weeks.
    :return: Number of microseconds.
    """
    return weeks * MICROSECONDS_IN_WEEK


def weeks_to_milliseconds(weeks: float) -> float:
    """
    Converts weeks to milliseconds.

    :param weeks: Number of weeks.
    :return: Number of milliseconds.
    """
    return weeks * MILLISECONDS_IN_WEEK


def weeks_to_seconds(weeks: float) -> float:
    """
    Converts weeks to seconds.

    :param weeks: Number of weeks.
    :return: Number of seconds.
    """
    return weeks * SECONDS_IN_WEEK


def weeks_to_minutes(weeks: float) -> float:
    """
    Converts weeks to minutes.

    :param weeks: Number of weeks.
    :return: Number of minutes.
    """
    return weeks * MINUTES_IN_WEEK


def weeks_to_hours(weeks: float) -> float:
    """
    Converts weeks to hours.

    :param weeks: Number of weeks.
    :return: Number of hours.
    """
    return weeks * HOURS_IN_WEEK


def weeks_to_days(weeks: float) -> float:
    """
    Converts weeks to days.

    :param weeks: Number of weeks.
    :return: Number of days.
    """
    return weeks * DAYS_IN_WEEK


def datetime_from(
    year: int, month: int, day: int, hour: int = 0, minute: int = 0, second: int = 0
) -> datetime:
    """
    Returns a datetime object in UTC based off the given parameters.

    :param year: The year.
    :param month: The month.
    :param day: The day.
    :param hour: The hour (optional, default=0).
    :param minute: The minute (optional, default=0).
    :param second: The second (optional, default=0).
    :return: An instance of a datetime.
    """
    return datetime(year, month, day, hour, minute, second)


def datetime_to_epoch_seconds_utc(date_time: datetime) -> float:
    """
    Given a datetime, return the number of seconds since the epoch UTC.

    :param date_time: An instance of a datetime.
    :return: A UTC second timestamp.
    """
    return (date_time - EPOCH).total_seconds()


def datetime_to_epoch_milliseconds_utc(date_time: datetime) -> float:
    """
    Given a datetime, return the number of milliseconds since the epoch UTC.

    :param date_time: An instance of a datetime.
    :return: A UTC millisecond timestamp.
    """
    epoch_seconds: float = datetime_to_epoch_seconds_utc(date_time)
    return seconds_to_milliseconds(epoch_seconds)


def datetime_to_epoch_microseconds_utc(date_time: datetime) -> float:
    """
    Given a datetime, return the number of microseconds since the epoch UTC.

    :param date_time: The datetime object to convert.
    :return: A UTC microsecond timestamp.
    """
    epoch_seconds: float = datetime_to_epoch_seconds_utc(date_time)
    return seconds_to_microseconds(epoch_seconds)


def datetime_from_epoch_seconds_utc(epoch_seconds_utc: float) -> datetime:
    """
    Given number of seconds since the epoch UTC, return a Python datetime object.

    :param epoch_seconds_utc: A UTC second timestamp.
    :return: A datetime object.
    """
    return EPOCH + timedelta(seconds=epoch_seconds_utc)


def datetime_from_epoch_milliseconds_utc(epoch_milliseconds_utc: float) -> datetime:
    """
    Given number of milliseconds since the epoch UTC, return a Python datetime object.

    :param epoch_milliseconds_utc: UTC millisecond timestamp.
    :return: Datetime object.
    """
    return EPOCH + timedelta(milliseconds=epoch_milliseconds_utc)


def datetime_from_epoch_microseconds_utc(epoch_microseconds_utc: float) -> datetime:
    """
    Given number of microseconds since the epoch UTC, return a Python datetime object.

    :param epoch_microseconds_utc: UTC microsecond timestamp.
    :return: A datetime object.
    """
    return EPOCH + timedelta(microseconds=epoch_microseconds_utc)


def datetimes_from_epoch_seconds_utc(epochs_seconds_utc: List[int]) -> List[datetime]:
    """
    Concerts a list of timestamps as seconds since the epoch UTC to a list of datetime objects.

    :param epochs_seconds_utc: List of timestamps to convert.
    :return: A list of datetimes.
    """
    return list(map(datetime_from_epoch_seconds_utc, epochs_seconds_utc))


def generate_timestamps_s_utc(
    start_timestamp_s_utc: float, sample_rate_hz: float, num_samples: int
) -> np.ndarray:
    """
    Given a starting timestamp, a sample rate, and a number of samples, compute timestamps for all samples.

    :param start_timestamp_s_utc: The start timestamp.
    :param sample_rate_hz: The sample rate.
    :param num_samples: The number of samples.
    :return: Timestamps for each sample.
    """
    sample_interval_s: float = 1.0 / sample_rate_hz
    return (np.arange(num_samples) * sample_interval_s) + start_timestamp_s_utc


def now() -> datetime:
    """
    Returns the current datetime in UTC.

    :return: The current datetime in UTC.
    """
    return datetime.utcnow()


class DateIterator:
    """
    This class provides an iterator over dates. That is, it takes a start date and an end date and returns a tuple
    of year, month, day for each date between the start and end.
    """

    def __init__(self, start_timestamp_utc_s: int, end_timestamp_utc_s: int):

        start_dt_full: datetime = datetime.utcfromtimestamp(start_timestamp_utc_s)
        end_dt_full: datetime = datetime.utcfromtimestamp(end_timestamp_utc_s)

        self.start_dt: datetime = datetime_from(
            start_dt_full.year, start_dt_full.month, start_dt_full.day
        )
        self.end_dt: datetime = datetime_from(
            end_dt_full.year, end_dt_full.month, end_dt_full.day
        )

        self._one_day: timedelta = timedelta(days=1)

    def __iter__(self) -> "DateIterator":
        return self

    def __next__(self) -> Tuple[str, str, str]:
        """
        Returns the next date in the iterator.
        :return: The next date in the iterator.
        """
        if self.start_dt > self.end_dt:
            raise StopIteration()

        year = str(self.start_dt.year)
        month = f"{self.start_dt.month:02}"
        day = f"{self.start_dt.day:02}"

        self.start_dt += self._one_day

        return year, month, day


class DateIteratorAPIM:
    """
    This class provides an iterator over dates. That is, it takes a start date and an end date and returns a tuple
    of year, month, day and hour for each date between the start and end.
    """

    def __init__(self, start_timestamp_utc_s: int, end_timestamp_utc_s: int):

        start_dt_full: datetime = datetime.utcfromtimestamp(start_timestamp_utc_s)
        end_dt_full: datetime = datetime.utcfromtimestamp(end_timestamp_utc_s)

        self.start_dt: datetime = datetime_from(
            start_dt_full.year,
            start_dt_full.month,
            start_dt_full.day,
            start_dt_full.hour,
        )
        self.end_dt: datetime = datetime_from(
            end_dt_full.year, end_dt_full.month, end_dt_full.day, end_dt_full.hour
        )

        self._one_day: timedelta = timedelta(days=1)
        self._one_hour: timedelta = timedelta(hours=1)

    def __iter__(self) -> "DateIteratorAPIM":
        return self

    def __next__(self) -> Tuple[str, str, str, str]:
        """
        Returns the next date in the iterator.

        :return: The next date in the iterator.
        """
        if self.start_dt > self.end_dt:
            raise StopIteration()

        year = str(self.start_dt.year)
        month = f"{self.start_dt.month:02}"
        day = f"{self.start_dt.day:02}"
        hour = f"{self.start_dt.hour:02}"

        self.start_dt += self._one_hour

        return year, month, day, hour


def truncate_dt_ymd(date_time: datetime) -> datetime:
    """
    Truncates the provides datetime to only year, month, day.

    :param date_time: The datetime to truncate.
    :return: A truncated datetime.
    """
    return datetime(date_time.year, date_time.month, date_time.day)


def truncate_dt_ymdh(date_time: datetime) -> datetime:
    """
    Truncates the provides datetime to only year, month, day, hour.

    :param date_time: The datetime to truncate.
    :return: A truncated datetime.
    """
    return datetime(date_time.year, date_time.month, date_time.day, date_time.hour)



"""
scales_dyadic.py
Physical to cyber conversion with preferred quasi-dyadic orders.
Compute the complete set of Nth order bands from Nyquist to the averaging frequency.
Compute the standard deviation of the Gaussian envelope for the averaging frequency.
Example: base10, base2, dB, bits

"""

import numpy as np
from typing import Tuple, Union

""" Smallest number for 64-, 32-, and 16-bit floats. Deploy to avoid division by zero or log zero singularities"""
EPSILON64 = np.finfo(np.float64).eps
EPSILON32 = np.finfo(np.float32).eps
EPSILON16 = np.finfo(np.float16).eps

# Convert microseconds to seconds
MICROS_TO_S = 1E-6

# 3*pi/4
M_OVER_N = 0.75 * np.pi

"""
Standardized scales
"""


class Slice:
    """
    Constants for slice calculations, supersedes inferno/slice
    """
    # Preferred Orders
    ORD1 = 1.  # Octaves; repeats nearly every three decades (1mHz, 1Hz, 1kHz)
    ORD3 = 3.  # 1/3 octaves; reduced temporal uncertainty for sharp transients and good for decade cycles
    ORD6 = 6.  # 1/6 octaves, good compromise for time-frequency resolition
    ORD12 = 12.  # Musical tones, continuous waves, long duration sweeps
    ORD24 = 24.  # High resolution; long duration widow, 24 bands per octave
    ORD48 = 48.  # Ultra-high spectral resolution for blowing minds and interstellar communications
    # Constant Q Base
    G2 = 2.  # Base two for perfect octaves and fractional octaves
    G3 = 10. ** 0.3  # Reconciles base2 and base10
    # Time
    T_PLANCK = 5.4E-44  # 2.**(-144)   Planck time in seconds
    T0S = 1E-42  # Universal Scale in S
    T1S = 1.    # 1 second
    T100S = 100.  # 1 hectosecond, IMS low band edge
    T1000S = 1000.  # 1 kiloseconds = 1 mHz
    T1M = 60.  # 1 minute in seconds
    T1H = T1M*60.  # 1 hour in seconds
    T1D = T1H*24.   # 1 day in seconds
    TU = 2.**58  # Estimated age of the known universe in seconds
    # Frequency
    F1HZ = 1.  # 1 Hz
    F1KHZ = 1_000.  # 1 kHz
    F0HZ = 1.E42  # 1/Universal Scale
    FU = 2.**-58  # 1/Estimated age of the known universe in s

    # NOMINAL SAMPLE RATES (REDVOX API M, 2022)
    FS1HZ = 1.        # SOH
    FS10HZ = 10.      # iOS Barometer
    FS30HZ = 30.      # Android barometer
    FS80HZ = 80.      # Infrasound
    FS200HZ = 200.    # Android Magnetometer, Fast
    FS400HZ = 400.    # Android Accelerometer and Gyroscope, Fast
    FS800HZ = 800.    # Infrasound and Low Audio
    FS8KHZ = 8_000.    # Speech Audio
    FS16KHZ = 16_000.  # ML Audio
    FS48KHZ = 48_000.  # Audio to Ultrasound


# DEFAULT CONSTANTS FOR TIME_FREQUENCY CANVAS
default_scale_base = Slice.G3
default_scale_order = Slice.ORD3
default_ref_frequency_hz = Slice.F1HZ

default_scale_order_min: float = 0.75  # Garces (2022)
default_fft_pow2_points_max: int = 2**15  # Computational FFT limit, tune to computing system
default_fft_pow2_points_min: int = 2**8  # For a tolerable display
default_mesh_pow2_pixels: int = 2**19  # Total of pixels per mesh, tune to plotting engine
default_time_display_s: float = 60.  # Physical time to display; sets display truncation


def scale_order_check(scale_order: float = default_scale_order):
    """
    Ensure no negative, complex, or unreasonably small orders are passed; override to 1/3 octave band
    Standard orders are scale_order = [1, 3, 6, 12, 24]. Order must be >= 0.75 or reverts to N=3
    :param scale_order: Band order,
    """
    # TODO: Refine
    # I'm confident there are better admissibility tests
    scale_order = np.abs(scale_order)  # Should be real, positive float
    if scale_order < default_scale_order_min:
        print('** Warning in scales_dyadic.scale_order_check')
        print('N<0.75 specified, overriding using N = ', default_scale_order_min)
        scale_order = default_scale_order_min
    return scale_order


def scale_multiplier(scale_order: float = default_scale_order):
    """
    Scale multiplier for scale bands of order N > 0.75
    :param scale_order: scale order
    :return:
    """
    scale_order = scale_order_check(scale_order)
    return M_OVER_N*scale_order


def cycles_from_order(scale_order: float) -> float:
    """
    Compute the number of cycles M for a specified band order N
    N is the quantization parameter for the constant Q wavelet filters

    :param scale_order: Band order, must be > 0.75 or reverts to N=0.75
    :return: cycles_M, number of cycled per normalized angular frequency
    """
    cycles_per_scale = scale_multiplier(scale_order)
    return cycles_per_scale


def order_from_cycles(cycles_per_scale: float) -> float:
    """
    Compute the number of cycles M for a specified band order N
    N is the quantization parameter for the constant Q wavelet filters

    :param cycles_per_scale: Should be greater than or equal than one
    :return: cycles_M, number of cycled per normalized angular frequency
    """
    # A single cycle is the min req
    if np.abs(cycles_per_scale) < 1:
        cycles_per_scale = 1.
    scale_order = cycles_per_scale/M_OVER_N
    # Order must be greater than min
    scale_order = scale_order_check(scale_order)

    return scale_order


def base_multiplier(scale_order: float = default_scale_order,
                    scale_base: float = default_scale_base):
    """
    Dyadic (log2) foundation for arbitrary base
    :param scale_order:
    :param scale_base:
    :return:
    """

    scale_order_check(scale_order)
    return scale_order/np.log2(scale_base)


def scale_from_frequency_hz(scale_order: float,
                            scale_frequency_center_hz: Union[np.ndarray, float],
                            frequency_sample_rate_hz: float) -> \
        Tuple[Union[np.ndarray, float], Union[np.ndarray, float]]:
    """
    Nondimensional scale and angular frequency for canonical Gabor/Morlet wavelet
    :param scale_frequency_center_hz: scale frequency in hz
    :param frequency_sample_rate_hz: sample rate in hz
    :return: scale_atom, scaled angular frequency
    """
    scale_angular_frequency = 2. * np.pi * scale_frequency_center_hz/frequency_sample_rate_hz
    scale_atom = cycles_from_order(scale_order)/scale_angular_frequency
    return scale_atom, scale_angular_frequency


def gaussian_sigma_from_frequency(frequency_sample_hz: float,
                                  frequency_hz: np.ndarray,
                                  scale_order: float = default_scale_order):
    """
    Standard deviation (sigma) of Gaussian envelope from frequency (Garces, 2023)
    Use 3/8 = 0.375
    :param frequency_hz: physical frequency in Hz
    :param frequency_sample_hz: sensor sample rate in Hz
    :param scale_order: Scale order
    :return: standard deviation from physical frequency in Hz
    """
    return 0.375*scale_order*frequency_sample_hz/frequency_hz


def scale_bands_from_ave(frequency_sample_hz: float,
                         frequency_ave_hz: float,
                         scale_order: float = default_scale_base,
                         scale_ref_hz: float = default_ref_frequency_hz,
                         scale_base: float = default_scale_base):
    """

    :param frequency_sample_hz:
    :param frequency_ave_hz:
    :param scale_order:
    :param scale_ref_hz:
    :param scale_base:
    :return: [band_nyq, band_ave, band_max, log2_ave_life_dyad]
    """

    # Framework constants
    scale_mult = scale_multiplier(scale_order)
    order_over_log2base = base_multiplier(scale_order, scale_base)

    # Dyadic transforms
    log2_scale_mult = np.log2(scale_mult)

    # Dependent on reference and averaging frequency
    log2_ave_physical = np.log2(scale_ref_hz/frequency_ave_hz)

    # Dependent on sensor sample rate
    log2_ref = np.log2(frequency_sample_hz/scale_ref_hz)
    log2_ave_cyber = np.log2(frequency_sample_hz/frequency_ave_hz)

    # Closest largest power of two to averaging frequency
    log2_ave_life_dyad = np.ceil(log2_scale_mult + log2_ave_cyber)

    # Shortest period, Nyquist limit
    band_nyq = int(np.ceil(order_over_log2base*(1-log2_ref)))
    # Closest period band
    band_ave = int(np.round(order_over_log2base*log2_ave_physical))
    # Longest period band
    band_max = int(np.floor(order_over_log2base*(log2_ave_life_dyad - log2_scale_mult - log2_ref)))

    return [band_nyq, band_ave, band_max, log2_ave_life_dyad]


def scale_band_from_frequency(frequency_input_hz: float,
                              scale_order: float = default_scale_base,
                              scale_ref_hz: float = default_ref_frequency_hz,
                              scale_base: float = default_scale_base):
    """
    For any one mid frequencies; not meant to be vectorized, only for comparing expectations.
    DOES NOT DEPEND ON SAMPLE RATE
    :param frequency_input_hz:
    :param scale_order:
    :param scale_ref_hz:
    :param scale_base:
    :return: [band_number, band_frequency_hz]
    """

    # TODO: Meant for single input values away from max and min bands
    # Framework constants
    order_over_log2base = base_multiplier(scale_order, scale_base)
    # Dependent on reference frequency
    log2_in_physical = np.log2(scale_ref_hz/frequency_input_hz)
    # Closest period band, TODO: check for duplicates if vectorized
    band_number = int(np.round(order_over_log2base*log2_in_physical))
    band_frequency_hz = scale_ref_hz*scale_base**(-band_number/scale_order)

    return [band_number, band_frequency_hz]


def scale_bands_from_pow2(frequency_sample_hz: float,
                          log2_ave_life_dyad: int,
                          scale_order: float = default_scale_base,
                          scale_ref_hz: float = default_ref_frequency_hz,
                          scale_base: float = default_scale_base):
    """

    :param frequency_sample_hz:
    :param log2_ave_life_dyad:
    :param scale_order:
    :param scale_ref_hz:
    :param scale_base:
    :return: [band_nyq, band_max]
    """

    # Framework constants
    scale_mult = scale_multiplier(scale_order)
    order_over_log2base = base_multiplier(scale_order, scale_base)

    # Dyadic transforms
    log2_scale_mult = np.log2(scale_mult)

    # Dependent on sensor sample rate
    log2_ref = np.log2(frequency_sample_hz/scale_ref_hz)

    # Shortest period, Nyquist limit
    band_nyq = int(np.ceil(order_over_log2base*(1-log2_ref)))
    # Longest period band
    band_max = int(np.floor(order_over_log2base*(log2_ave_life_dyad - log2_scale_mult - log2_ref)))

    return [band_nyq, band_max]


def period_from_bands(band_min: int,
                      band_max: int,
                      scale_order: float = default_scale_base,
                      scale_ref_hz: float = default_ref_frequency_hz,
                      scale_base: float = default_scale_base):

    bands = np.arange(band_min, band_max+1)
    # Increasing order
    period = scale_base**(bands/scale_order)/scale_ref_hz
    return period


def frequency_from_bands(band_min: int,
                         band_max: int,
                         scale_order: float = default_scale_base,
                         scale_ref_hz: float = default_ref_frequency_hz,
                         scale_base: float = default_scale_base):

    bands = np.arange(band_min, band_max+1)
    # Flip so it increases
    frequency = np.flip(scale_ref_hz*scale_base**(-bands/scale_order))
    return frequency


def log_frequency_hz_from_fft_points(
        frequency_sample_hz: float,
        fft_points: int,
        scale_order: float = default_scale_base,
        scale_ref_hz: float = default_ref_frequency_hz,
        scale_base: float = default_scale_base
) -> np.ndarray:

    # TODO: Make function to round to to nearest power of two and perform all-around error checking for pow2
    # See log 2 functions below
    log2_ave_life_dyad = int(np.ceil(np.log2(fft_points)))
    # Framework constants
    scale_mult = scale_multiplier(scale_order)
    order_over_log2base = base_multiplier(scale_order, scale_base)

    # Dyadic transforms
    log2_scale_mult = np.log2(scale_mult)

    # Dependent on sensor sample rate
    log2_ref = np.log2(frequency_sample_hz/scale_ref_hz)

    # Shortest period, Nyquist limit
    band_nyq = int(np.ceil(order_over_log2base*(1-log2_ref)))
    # Shortest period, nominal 8/10 of Nyquist
    band_aa = int(np.ceil(order_over_log2base*(np.log2(2.5) - log2_ref)))
    # Longest period band
    band_max = int(np.floor(order_over_log2base*(log2_ave_life_dyad - log2_scale_mult - log2_ref)))

    # Stopped before Nyquist, before max
    # bands_old = np.arange(band_nyq+1, band_max+1)
    # Stopped at 0.8 of Nyquist
    bands = np.arange(band_aa, band_max+1)
    # Flip so frequency increases up to one band below Nyquist
    frequency_logarithmic_hz = np.flip(scale_ref_hz*scale_base**(-bands/scale_order))
    return frequency_logarithmic_hz


def lin_frequency_hz_from_fft_points(
        frequency_sample_hz: float,
        fft_points: int,
        scale_order: float = default_scale_base,
        scale_ref_hz: float = default_ref_frequency_hz,
        scale_base: float = default_scale_base
) -> np.ndarray:
    """

    :param frequency_sample_hz:
    :param fft_points:
    :param scale_order:
    :param scale_ref_hz:
    :param scale_base:
    :return:
    """

    frequency_log_hz = log_frequency_hz_from_fft_points(frequency_sample_hz, fft_points,
                                                        scale_order, scale_ref_hz, scale_base)
    frequency_lin_min = frequency_log_hz[0]  # First band
    frequency_lin_max = frequency_log_hz[-1]  # Last band
    frequency_linear_hz = np.linspace(start=frequency_lin_min, stop=frequency_lin_max, num=len(frequency_log_hz))
    return frequency_linear_hz


# TODO: Turn into a test function
if __name__ == "__main__":
    # Framework specs
    scale_order0 = 6.
    scale_base0 = Slice.G3
    scale_ref0 = Slice.F1HZ

    # Sensor specific
    frequency_sample_hz0 = 100.
    print('\nNominal Sample Rate, Hz:', frequency_sample_hz0)

    fft_points_max = default_fft_pow2_points_max
    fft_points_min = default_fft_pow2_points_min
    max_mesh_pixels = default_mesh_pow2_pixels
    # Mesh time display limit
    time_display_s = default_time_display_s
    # TODO: Minimum number of points in display

    time_display_points_float: float = time_display_s*frequency_sample_hz0
    time_display_points_pow2: int = int(2**np.ceil(np.log2(time_display_points_float)))
    print('Display time, s:', time_display_s)
    print('Display points:', time_display_points_float)
    print('Display points pow2:', time_display_points_pow2)
    print('Display pow2 duration:', time_display_points_pow2/frequency_sample_hz0)

    # Works for 60s display window - display window becomes the max averaging lifetime
    # We have a computational averaging lifetime (~2^16) and a max averaging lifetime (display window)
    if time_display_points_pow2 > fft_points_max:
        fft_points0 = fft_points_max
    else:
        fft_points0 = time_display_points_pow2

    # Do not drop below stft_fft_points min
    if fft_points0 < fft_points_min:
        # 256 points is minimal
        fft_points0 = fft_points_min

    print('MAX FFT: Order, Base, Reference, log2(FFT points)')
    print(scale_order0, scale_base0, scale_ref0, np.log2(fft_points0))

    physical_frequency_hz = \
        log_frequency_hz_from_fft_points(frequency_sample_hz0, fft_points0,
                                         scale_order0, scale_ref0, scale_base0)
    number_of_frequencies = len(physical_frequency_hz)
    reduction_factor_float = number_of_frequencies*fft_points0/max_mesh_pixels

    if reduction_factor_float > 1:
        reduction_factor_pow2 = 2**np.ceil(np.log2(reduction_factor_float))
    else:
        reduction_factor_pow2 = 1.

    print('Physical frequency, Min, Max:', physical_frequency_hz[0], physical_frequency_hz[-1])
    print('Averaging window duration:', fft_points0/frequency_sample_hz0)
    print('Number of bands:', number_of_frequencies)
    print('Mesh reduction factor:', reduction_factor_pow2)







"""
This module constructs synthetic signals
"""

import numpy as np
import scipy.signal as signal
from scipy.integrate import cumulative_trapezoid
from typing import Optional, Tuple, Union
from libquantum import utils, scales, atoms


def gabor_loose_grain(band_order_Nth: float,
                      number_points: int,
                      scale_frequency_center_hz: float,
                      frequency_sample_rate_hz: float,
                      index_shift: float = 0,
                      frequency_base_input: float = scales.Slice.G2) -> Tuple[np.ndarray, np.ndarray, float]:
    """
    Loose grain with tight Tukey wrap to ensure zero at edges

    :param band_order_Nth: Nth order of constant Q bands
    :param number_points: Number of points in the signal
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate of frequency in Hz
    :param index_shift: index of shift for the Gabor chirp, default of zero
    :param frequency_base_input: G2 or G3. Default is G2
    :return: numpy array with Tukey grain
    """

    # Fundamental chirp parameters
    cycles_M, quality_factor_Q, gamma = \
        atoms.chirp_MQG_from_N(band_order_Nth, index_shift, frequency_base_input)
    scale_atom = atoms.chirp_scale(cycles_M, scale_frequency_center_hz, frequency_sample_rate_hz)
    p_complex = atoms.chirp_p_complex(scale_atom, gamma, index_shift)

    # # Time from nominal duration
    # grain_duration_s = cycles_M/scale_frequency_center_hz
    # Time from number of points
    time_s = np.arange(number_points)/frequency_sample_rate_hz
    offset_time_s = np.max(time_s)/2.

    xtime_shifted = atoms.chirp_time(time_s, offset_time_s, frequency_sample_rate_hz)
    wavelet_gauss = np.exp(-p_complex * xtime_shifted**2)
    wavelet_gabor = wavelet_gauss * np.exp(1j * cycles_M*xtime_shifted/scale_atom)
    loose_grain_taper = utils.taper_tukey(wavelet_gabor, 0.1)
    loose_grain = np.copy(wavelet_gabor) * loose_grain_taper

    return loose_grain, time_s, scale_atom


def gabor_tight_grain(band_order_Nth: float,
                      scale_frequency_center_hz: float,
                      frequency_sample_rate_hz: float,
                      index_shift: float = 0,
                      frequency_base_input: float = scales.Slice.G2) -> np.ndarray:
    """
    Gabor grain with tight Tukey wrap to ensure zero at edges

    :param band_order_Nth: Nth order of constant Q bands
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate of frequency in Hz
    :param index_shift: index of shift
    :param frequency_base_input: G2 or G3. Default is G2
    :return: numpy array with Tukey grain
    """

    # Fundamental chirp parameters
    cycles_M, quality_factor_Q, gamma = \
        atoms.chirp_MQG_from_N(band_order_Nth, index_shift, frequency_base_input)
    scale_atom = atoms.chirp_scale(cycles_M, scale_frequency_center_hz, frequency_sample_rate_hz)
    p_complex = atoms.chirp_p_complex(scale_atom, gamma, index_shift)

    # Time from nominal duration
    grain_duration_s = cycles_M/scale_frequency_center_hz
    time_s = np.arange(int(np.round(grain_duration_s*frequency_sample_rate_hz)))/frequency_sample_rate_hz
    offset_time_s = np.max(time_s)/2.

    xtime_shifted = atoms.chirp_time(time_s, offset_time_s, frequency_sample_rate_hz)
    wavelet_gauss = np.exp(-p_complex * xtime_shifted**2)
    wavelet_gabor = wavelet_gauss * np.exp(1j * cycles_M*xtime_shifted/scale_atom)
    tight_grain_taper = utils.taper_tukey(wavelet_gabor, 0.1)
    tight_grain = np.copy(wavelet_gabor)*tight_grain_taper

    return tight_grain


def tukey_tight_grain(band_order_Nth: float,
                      scale_frequency_center_hz: float,
                      frequency_sample_rate_hz: float,
                      fraction_cosine: float = 0.5,
                      index_shift: float = 0,
                      frequency_base_input: float = scales.Slice.G2) -> np.ndarray:
    """
    Tukey grain with same support as Gabor atom

    :param band_order_Nth: Nth order of constant Q bands
    :param scale_frequency_center_hz: center frequency fc in Hz
    :param frequency_sample_rate_hz: sample rate of frequency in Hz
    :param fraction_cosine: fraction of the window inside the cosine tapered window, shared between the head and tail.
        Default is 0.5
    :param index_shift: index of shift
    :param frequency_base_input: G2 or G3. Default is G2
    :return: numpy array with Tukey grain
    """

    # Fundamental chirp parameters
    cycles_M, quality_factor_Q, gamma = \
        atoms.chirp_MQG_from_N(band_order_Nth, index_shift, frequency_base_input)
    scale_atom = atoms.chirp_scale(cycles_M, scale_frequency_center_hz, frequency_sample_rate_hz)
    p_complex = atoms.chirp_p_complex(scale_atom, gamma, index_shift)

    # Time from nominal duration
    grain_duration_s = cycles_M/scale_frequency_center_hz
    time_s = np.arange(int(np.round(grain_duration_s*frequency_sample_rate_hz)))/frequency_sample_rate_hz
    offset_time_s = np.max(time_s)/2.

    xtime_shifted = atoms.chirp_time(time_s, offset_time_s, frequency_sample_rate_hz)
    # Pull out phase component from gaussian envelope
    wavelet_gauss_phase = np.imag(-p_complex * xtime_shifted**2)
    wavelet_gabor = np.exp(1j * cycles_M*xtime_shifted/scale_atom + 1j * wavelet_gauss_phase)
    tight_grain_taper = utils.taper_tukey(wavelet_gabor, fraction_cosine)
    tight_grain = np.copy(wavelet_gabor)*tight_grain_taper

    return tight_grain


def gabor_grain_frequencies(frequency_order_input: float,
                            frequency_low_input: float,
                            frequency_high_input: float,
                            frequency_sample_rate_input: float,
                            frequency_base_input: float = scales.Slice.G2,
                            frequency_ref_input: float = 1.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Frequencies for g-chirps

    :param frequency_order_input: Nth order
    :param frequency_low_input: lowest frequency of interest
    :param frequency_high_input: highest frequency of interest
    :param frequency_sample_rate_input: sample rate
    :param frequency_base_input: G2 or G3. Default is G2
    :param frequency_ref_input: reference frequency. Default is 1.0
    :return: three numpy arrays with center frequency, start frequency and end frequency
    """

    scale_order, scale_base, _, frequency_ref, frequency_center_algebraic, \
    frequency_center, frequency_start, frequency_end = \
        scales.band_frequencies_low_high(frequency_order_input, frequency_base_input,
                                         frequency_ref_input,
                                         frequency_low_input,
                                         frequency_high_input,
                                         frequency_sample_rate_input)

    return frequency_center, frequency_start, frequency_end


def chirp_rdvxm_noise_16bit(duration_points: int = 2**12,
                            sample_rate_hz: float = 80.,
                            noise_std_loss_bits: float = 4.,
                            frequency_center_hz: Optional[float] = None):
    """
    Construct chirp with linear frequency sweep, white noise added, anti-aliased filter applied

    :param duration_points: number of points, length of signal. Default is 2 ** 12
    :param sample_rate_hz: sample rate in Hz. Default is 80.0
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :param frequency_center_hz: center frequency fc in Hz. Optional
    :return: numpy ndarray with anti-aliased chirp with white noise
    """

    duration_s = duration_points/sample_rate_hz
    if frequency_center_hz:
        frequency_start_hz = 0.5*frequency_center_hz
        frequency_end_hz = sample_rate_hz/4.
    else:
        frequency_center_hz = 8./duration_s
        frequency_start_hz = 0.5*frequency_center_hz
        frequency_end_hz = sample_rate_hz/4.

    sig_time_s = np.arange(int(duration_points))/sample_rate_hz
    chirp_wf = signal.chirp(sig_time_s, frequency_start_hz, sig_time_s[-1],
                            frequency_end_hz, method='linear', phi=0, vertex_zero=True)
    chirp_wf *= taper_tukey(chirp_wf, 0.25)
    noise_wf = white_noise_fbits(sig=chirp_wf, std_bit_loss=noise_std_loss_bits)
    chirp_white = chirp_wf + noise_wf
    chirp_white_aa = antialias_halfNyquist(chirp_white)
    chirp_white_aa.astype(np.float16)

    return chirp_white_aa


def sawtooth_rdvxm_noise_16bit(duration_points: int = 2**12,
                               sample_rate_hz: float = 80.,
                               noise_std_loss_bits: float = 4.,
                               frequency_center_hz: Optional[float] = None) -> np.ndarray:
    """
    Construct an anti-aliased sawtooth waveform with white noise

    :param duration_points: number of points, length of signal. Default is 2 ** 12
    :param sample_rate_hz: sample rate in Hz. Default is 80.0
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :param frequency_center_hz: center frequency fc in Hz. Optional
    :return: numpy ndarray with anti-aliased sawtooth signal with white noise
    """

    duration_s = duration_points/sample_rate_hz
    if frequency_center_hz:
        frequency_center_angular = 2*np.pi*frequency_center_hz
    else:
        frequency_center_hz = 8./duration_s
        frequency_center_angular = 2*np.pi*frequency_center_hz

    sig_time_s = np.arange(int(duration_points))/sample_rate_hz
    saw_wf = signal.sawtooth(frequency_center_angular*sig_time_s, width=0)
    saw_wf *= taper_tukey(saw_wf, 0.25)
    noise_wf = white_noise_fbits(sig=saw_wf, std_bit_loss=noise_std_loss_bits)
    saw_white = saw_wf + noise_wf
    saw_white_aa = antialias_halfNyquist(saw_white)
    saw_white_aa.astype(np.float16)

    return saw_white_aa


def sawtooth_doppler_noise_16bit(phase_radians: np.ndarray,
                                 noise_std_loss_bits: float = 4.) -> np.ndarray:
    """
    Construct an anti-aliased sawtooth waveform with white noise

    :param phase_radians: time-varying phase in radians
    :param noise_std_loss_bits: number of bits below signal standard deviation. Default is 4.0
    :return: numpy ndarray with anti-aliased sawtooth signal with white noise
    """

    saw_wf = signal.sawtooth(phase_radians, width=0)
    saw_wf *= taper_tukey(saw_wf, 0.25)
    noise_wf = white_noise_fbits(sig=saw_wf, std_bit_loss=noise_std_loss_bits)
    saw_white = saw_wf + noise_wf
    saw_white_aa = antialias_halfNyquist(saw_white)
    saw_white_aa.astype(np.float16)

    return saw_white_aa


def chirp_linear_in_noise(snr_bits: float,
                          sample_rate_hz: float,
                          duration_s: float,
                          frequency_start_hz: float,
                          frequency_end_hz: float,
                          intro_s: Union[int, float],
                          outro_s: Union[int, float]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Construct chirp with linear frequency sweep, white noise added.

    :param snr_bits: number of bits below signal standard deviation
    :param sample_rate_hz: sample rate in Hz
    :param duration_s: duration of chirp in seconds
    :param frequency_start_hz: start frequency in Hz
    :param frequency_end_hz: end frequency in Hz
    :param intro_s: number of seconds before chirp
    :param outro_s: number of seconds after chirp
    :return: numpy ndarray with waveform, numpy ndarray with time in seconds
    """

    sig_time_s = np.arange(int(sample_rate_hz*duration_s))/sample_rate_hz
    chirp_wf = signal.chirp(sig_time_s, frequency_start_hz, sig_time_s[-1],
                            frequency_end_hz, method='linear', phi=0, vertex_zero=True)
    chirp_wf *= taper_tukey(chirp_wf, 0.25)
    sig_wf = np.concatenate((np.zeros(int(intro_s*sample_rate_hz)),
                             chirp_wf,
                             np.zeros(int(outro_s*sample_rate_hz))))
    noise_wf = white_noise_fbits(sig=sig_wf, std_bit_loss=snr_bits)
    synth_wf = sig_wf+noise_wf
    synth_time_s = np.arange(len(synth_wf))/sample_rate_hz
    return synth_wf, synth_time_s


def white_noise_fbits(sig: np.ndarray,
                      std_bit_loss: float) -> np.ndarray:
    """
    Compute white noise with zero mean and standard deviation that is snr_bits below the input signal

    :param sig: input signal, detrended
    :param std_bit_loss: number of bits below signal standard deviation
    :return: gaussian noise with zero mean
    """
    sig_std = np.std(sig)
    # This is in power, or variance
    noise_loss = 2.**std_bit_loss
    std_from_fbits = sig_std/noise_loss
    # White noise, zero mean
    sig_noise = np.random.normal(0, std_from_fbits, size=sig.size)
    return sig_noise


def taper_tukey(sig_or_time: np.ndarray,
                fraction_cosine: float) -> np.ndarray:
    """
    Constructs a symmetric Tukey window with the same dimensions as a time or signal numpy array.
    fraction_cosine = 0 is a rectangular window, 1 is a Hann window

    :param sig_or_time: input signal or time
    :param fraction_cosine: fraction of the window inside the cosine tapered window, shared between the head and tail
    :return: tukey taper window amplitude
    """
    number_points = np.size(sig_or_time)
    amplitude = signal.windows.tukey(M=number_points, alpha=fraction_cosine, sym=True)
    return amplitude


def antialias_halfNyquist(synth: np.ndarray) -> np.ndarray:
    """
    Anti-aliasing filter with -3dB at 1/4 of sample rate, 1/2 of Nyquist

    :param synth: array with signal data
    :return: numpy array with anti-aliased signal
    """
    # Anti-aliasing filter with -3dB at 1/4 of sample rate, 1/2 of Nyquist
    # Signal frequencies are scaled by Nyquist
    filter_order = 2
    edge_high = 0.5
    [b, a] = signal.butter(filter_order, edge_high, btype='lowpass')
    synth_anti_aliased = signal.filtfilt(b, a, np.copy(synth))
    return synth_anti_aliased


def frequency_algebraic_Nth(frequency_geometric: np.ndarray,
                            band_order_Nth: float) -> np.ndarray:
    """
    Compute algebraic frequencies in band order

    :param frequency_geometric: geometric frequencies
    :param band_order_Nth:  Nth order of constant Q bands
    :return:
    """
    frequency_algebra = frequency_geometric*(np.sqrt(1+1/(8*band_order_Nth**2)))
    return frequency_algebra


def integrate_cumtrapz(timestamps_s: np.ndarray,
                       sensor_wf: np.ndarray,
                       initial_value: float = 0) -> np.ndarray:
    """
    cumulative trapezoid integration using scipy.integrate.cumulative_trapezoid

    :param timestamps_s: timestamps corresponding to the data in seconds
    :param sensor_wf: data to integrate using cumulative trapezoid
    :param initial_value: the value to add in the initial of the integrated data to match length of input (default is 0)
    :return: integrated data with the same length as the input
    """
    integrated_data = cumulative_trapezoid(x=timestamps_s,
                                           y=sensor_wf,
                                           initial=initial_value)
    return integrated_data
